!+
! Module: eexpmap_mod
!
! Authors:  Hermann Brunner - AIP
!           Alexis Finoguenov - AIP   
!           Georg Lamer - AIP       
!          
! Description: EPIC exposure maps for use with the EPIC detection  tasks
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module eexpmap_mod

  use cal
  use oal
  use dal
  use param  
  use dss
  use edetect_io_exp
  use pix2map_mod
  use errorHandling
!  use attcalc_mod
!  use emutils_mod
 
  implicit none 
     
  public eexpmap

  real(kind=double), dimension(:,:), allocatable :: fexpext
  type :: start_stopT
    real(kind=double), dimension(:), pointer :: start, stop
  end type


contains 

  !---------------------------------------------------------------------
  subroutine eexpmap
  !---------------------------------------------------------------------
    real(kind=single), dimension (:), allocatable :: amplo, amphi
    real(kind=double), dimension(:), pointer :: theta, phi
!!!!    real(kind=single), dimension(:), pointer :: theta, phi
    real(kind=double) :: dtheta, dphi, eps, ctime, hducut
    !real(kind=double), parameter :: pi = 3.141592654, pix = 4.0, &
    !     flngth=7500.0, fac = pix/3600.0 * (pi/180.0) * flngth, &
    !                                factheta = pix/fac
    real(kind=double), parameter :: pi = 3.141592654, pix = 4.0 
    real(kind=double) :: flngth, fac
    real(kind=double), dimension(:), pointer :: time, gstart, gstop
    real(kind=double), dimension (:), allocatable :: E, area, yextra
    real(kind=single), dimension(:), allocatable :: transmission, &
                                     effarea_cen, effgrad, teffgrad 
 
    real(kind=single) :: expos
    real(kind=single), dimension(:,:), allocatable :: &
                                           transmission_save, area_save   
    integer(kind=int16), dimension(0:600) :: rawX, rawY
    real(kind=double), dimension(:), pointer :: ra0, de0, pa0
    integer(kind=int16), dimension(:), pointer :: bbrawx, bbrawy,bbye,bbox,bbofs,bboffx,bboffy, & 
                                                  bbadflag,bbadtype
    integer(kind=int16), dimension(:), allocatable :: badpixtype
    integer(kind=int8),  dimension(:),  pointer :: bbnr 
    integer(kind=int32) :: i, k, imask, ngti_ext, & 
                           naxis1, naxis2, j, &
                           ncoras, irange, itfilt, itfilt1, ngti1

! extra working parameters. Alexis.
    integer(kind=int32) :: npx0, npx1, npy0, npy1, mpx0, mpx1, mpy0, mpy1 
    integer(kind=int32) :: iexp,jexp,iccd,nccd,nccd0,ic0,ic1
    integer(kind=int16) :: ipx,ipy,ipyp,ipym2
    integer(kind=int16), parameter :: ipmin = 10
!    integer(kind=int32) :: ngrade, igrade, ntgrade, maxgrad, nsubmode, nfilter, nband, nmask
    integer(kind=int32) :: nsubmode, nfilter, nband, nmask
!    integer(kind=int32), dimension(0:100) :: grade, tgrade
!    real(kind=single), dimension(0:100) :: effg
    logical :: useDetector, notsaid,  rndmz, useImageCS, &
               usedss,  qbb, qofs, qofsy, withVig, fastmode, useoffset, offsetflag

!---------------------------------------------------------------------
! extra parameters for reading DLIMAP. JR
!---------------------------------------------------------------------
    integer, parameter :: nccd1 = 12
    type(ColumnT) :: dliexpco1
    type(TableT) :: tabdliexpco
    logical ::  hasdliset0
    integer :: ia, ndli
    real(kind=single), dimension(:), pointer :: bdliexpco1
    character(len=20) :: name1
    character(len=6) :: xxname1
    logical :: dlimode
!---------------------------------------------------------------------
!---------------------------------------------------------------------


! extra input switches
    real(kind=double) :: detxpx2mm_save,detypx2mm_save,detxcrvl_save, detycrvl_save
!    real(kind=single) :: detxpx2mm,detypx2mm,detxcrpx, detycrpx,&
!         & detxcrvl, detycrvl
    real(kind=double) :: detxpx2mm,detypx2mm,detxcrpx, detycrpx,&
         & detxcrvl, detycrvl
    real(kind=single) :: x0,y0,x1,y1,xsp,xsn,ysp,ysn,fhx,fhy,fh2,wgh
    real(kind=single) :: xn,xm,xp,yn,ym,yp,rx,ry,rx0,ry0,pixfac
!   real(kind=single) :: pixoff
    real(kind=single) :: wmm,wpp,wmp,wnp,wnm,wnn,wmn,wpn,wpm
    integer :: nsplit, lx, ly
    integer, dimension(:), pointer :: shape
!    integer(kind=int8), dimension(:,:,:), pointer :: patterns
    real(kind=double), dimension(:), pointer :: Xmm, Ymm, Zmm
!!!!    real(kind=single), dimension(:), pointer :: Xmm, Ymm, Zmm
    real(kind=single), dimension(0:3) :: Xmm1, Ymm1, Zmm1
    integer :: ix,iy
!    real(kind=single), dimension(:), pointer :: XmmSc, YmmSc, ZmmSc
    character(len=20) :: attinstr,name
    character(len=6) :: bbname, ocname
!   merge_gti characrer(len=12) :: ctype1, ctype2
    character(len=1024) :: stattit, gtiblock
!    integer(kind=int32) :: diff0order, ipxd2, bbrawy0, bbrawy1, nbb0, nbb1
    integer(kind=int32) :: diff0order, ipxd2, bbrawy0, bbrawy1

    real(kind=double), dimension(:,:), pointer :: Cmatrix
    real(kind=single), dimension(:,:,:), allocatable :: exp0 
    real(kind=single), dimension(:,:,:,:), allocatable :: exp1
    real(kind=double), dimension(0:gtimax) :: fstart, fstop, tempstart, tempstop
    integer(kind=int32), dimension(0:100) :: flagmin, flagmax
    integer(kind=int32) :: flagmintemp,flagmaxtemp,nflagranges
    integer(kind=int32), dimension(0:100) :: iccd0, iccd1
    integer(kind=int32), dimension(0:100,0:100) :: flagmask
    integer(kind=int32), dimension(:), pointer :: onBits, offBits
    integer(kind=int32), dimension(1:100) :: flagmargins,flagmargins2
    

    real(kind=double), parameter  :: radian = 57.29577951E0    
    real(kind=double), dimension(0:0), target :: tstart, tstop
!    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsize
    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsizex, skypixsizey
    integer(kind=int32) :: natt,  iblock, ntgti, ntempgti
    integer(kind=int32), dimension(12) :: winx0, winy0, windx, windy
    integer(kind=int32), parameter :: nattmax=10000
    integer(kind=int32), parameter :: nextmax=12
    real(kind=double), dimension(nattmax) :: xnew, ynew, crnew, srnew
    integer(kind=int32), dimension(12) :: ngti, acomp
    real(kind=single) :: xsky, ysky
    real(kind=single), dimension(0:5) :: xsky4, ysky4
    real(kind=single) :: fk, bix, biy
    real(kind=single), dimension(:), allocatable :: wght 
    real(kind=double) :: bmtrx10, bmtrx11, bmtrx20, bmtrx21
    integer(kind=int16), dimension(:,:), pointer :: map
    character(len=7) :: datamode, filterid
    character(len=20) :: submode                          
!    character(len=255) :: wstr
! Pixel size parameters from EPN_LINCOORD_0001.CCF
    ! real(kind=single), parameter :: MM_PX_X=0.15, MM_PX_Y=0.15, MM_PX_LY=0.2
    ! real(kind=single), parameter :: MOS_PX=0.04
    real(kind=single) :: MM_PX_X, MM_PX_Y, MM_PX_LY, MOS_PX
    real(kind=single) :: xhsz_mm, yhsz_mm, xexpu, xexpl, yexpu, yexpl, fracx,&
         fracy, frac,rbbc,xshift,yshift,xcen,ycen,exptest
    integer(kind=int16) :: imod
    integer(kind=int16) :: iexp0,iexp1,jexp0,jexp1
    integer :: icomp, ncomp, ifilt, nfilt, ftype, ntest
    character(len=80) :: fname,binType

    integer(kind=int32) :: ibb, nbb,nofs,iofs,ibbc,ipxad,ipyad,ibbc1,irpx2,irbbc2
    integer(kind=int32) :: margin, margin2, bordermargin, offsetmargin
    logical :: pos, mgti, hasgti, hasgtin, close, closeos, hasoffset0, haswcs2 
    logical, dimension(nextmax) :: relgti

    logical, dimension(:), allocatable :: qalpha, q45
    real(kind=single), dimension(:), allocatable :: alsin, alcos, hs
    real(kind=double) :: alpha, dx, dy
    real(kind=single) :: ootfrac

    integer(int16), dimension(0:0)             :: opticsx, opticsy ! RDS
    real(double), dimension(:), pointer        :: opticsx_cc, opticsy_cc, &
                                                  opticsz_cc    ! RDS

    type(start_stopT), dimension(nextmax) :: gti
    type(start_stopT) :: tempgti
    type(DataSetT) :: set1,set2,set3
    type(DataSetT), dimension(:), allocatable :: setO
    type(ArrayT) :: arr0
    type(ArrayT), dimension(:), allocatable :: arr
    type(TableT) :: tabcoras,tabgti,taboffs
    type(TableT), dimension(nextmax) :: tabi
    type(ColumnT) :: coltime,coldeo,colrao,colpa,crawx,crawy,cye, &
                     colgtista,colgtisto,cox,cofs,coffx,coffy,cnr,cbadflag,cbadtype
    type(ColumnT), dimension(nextmax) :: colstart,colstop       
    type(BlockT) :: dblock
    type(DssT) :: dss
    !AI: Arrays not needed. Let's try as scalar, this could help 
    !the execution in solaris...
    !type(DScompT), dimension(0:100) :: dssc 
    !type(DSfilterT), dimension(0:100,0:100) :: dssf 
    type(DScompT) :: dssc 
    type(DSfilterT) :: dssf 
    type(RangeT), dimension(:), allocatable:: frange
    type(BoundT) :: lbound,ubound
	integer :: nSize

    real(kind=double) :: rfac, rfac2x, rfac2y


    nullify (theta, phi)
    nullify (Xmm, Ymm, Zmm)
    nullify (opticsx_cc, opticsy_cc, opticsz_cc)
    nullify (Cmatrix, map)

    !!! IT 2017/06 nullify more pointers if compiling with gfortran
    nullify (bbrawx, bbrawy, bboffx, bboffy)
    nullify (bbye, bbox, bbofs)
    nullify (bbadflag,bbadtype)


    !type :: ExpoT
    !  real(kind=single), dimension(:,:), pointer :: exp
    !end type
    !type(ExpoT), dimension(:), allocatable :: earr
    
    flngth = CAL_getMiscellaneousDataValue('FOCAL_LENGTH')
    call CAL_setState(EPN);
    MM_PX_X = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')
    MM_PX_Y = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y')
    MM_PX_LY = CAL_getMiscellaneousDataValue('LINCOORD:LINCOORD%MM_PX_LY')
    call CAL_setState(EMOS1);
    MOS_PX = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')

    fac = pix/3600.0 * (pi/180.0) * flngth

! these will be read from user
!    UserPatterns =  booleanParameter("withpatterns")
    useDetector =  booleanParameter("withdetcoords")
!    useDSS = booleanParameter("usedss")
    withVig = booleanParameter("withvignetting")
    fastmode = booleanParameter("usefastpixelization")    
!------------------------------------------------------------------
! New parameter to allow/or not the use the MIP correction - (JR)
    dlimode = booleanParameter("usedlimap")
!------------------------------------------------------------------

! remove useoffset and hducut parameters
! 
!    useoffset = booleanParameter("useoffset")    
!    if(useoffset) then
!       hducut=realParameter("hducut")
!    endif

!   Change useoffset to 'yes', check DSS filters whether offset columns  
!   are excluded (GL 15/08/06) 

    useDSS=.true.
    useoffset=.true.
    hducut=0.0

    useImageCS = .true.
    rndmz = .false.
    diff0order = 0
    eps = realParameter("attrebin") / 3600.0
! Changed (23/01/07 GL):
!    ibbc = intParameter("badclean")
     ibbc = 0


    nband=parameterCount("pimin")
    allocate (earr(0:nband-1))      
    allocate (setO(0:nband-1))
    allocate (arr(0:nband-1))
    allocate (e(0:nband-1))
    allocate (area(0:nband-1))
    allocate (transmission(0:nband-1))
    allocate (effgrad(0:nband-1))
    allocate (teffgrad(0:nband-1))
    allocate (amplo(0:nband-1))
    allocate (amphi(0:nband-1))
    allocate (effarea_cen(0:nband-1))
    allocate (area_save(0:300,0:nband))
    allocate (transmission_save(0:300,0:nband))
    allocate (wght(0:nband-1))
    allocate (yextra(0:700))

    do i=0,nband-1 
      amplo(i) = intParameter("pimin",i)
      amphi(i) = intParameter("pimax",i)
    end do
    
! input of attitude history file

    stattit=stringParameter("attitudeset")
    set2=dataset(stattit,Read)
    tabcoras = table(set2,"ATTHK")
    ncoras = numberOfRows(tabcoras)
    coltime = column(tabcoras,"TIME", Read)    
    time => real64Data(coltime)
    colrao = column(tabcoras,"AHFRA", Read)
    ra0 => real64Data(colrao)
    coldeo = column(tabcoras,"AHFDEC", Read)
    de0 => real64Data(coldeo)
    colpa = column(tabcoras,"AHFPA", Read)
    pa0 => real64Data(colpa)

! input of EPIC FITS image
    set1=dataset(stringParameter("imageset"),Read)
    arr0 = array(set1, "PRIMARY", Read)
    if (hasAttribute(set1,"INSTRUME")) then  
      attinstr  = stringAttribute(set1,"INSTRUME")
    else
      attinstr  = stringAttribute(arr0,"INSTRUME")
    end if      

!   Now that we know what the instrument is, setState immediately
!   to get all the instrument-specific parameters right.
    if (attinstr == "EPN") then
      call CAL_setState(EPN)
    else if (attinstr == "EMOS1") then
      call CAL_setState(EMOS1)
    else if (attinstr == "EMOS2") then
      call CAL_setState(EMOS2)
    end if
!   Specifically, recalculate the quantities flngth and fac, which were
!   previously calculated w/o calling setState, and are thus wrong.
    flngth = CAL_getMiscellaneousDataValue('FOCAL_LENGTH')
    fac = pix/3600.0 * (pi/180.0) * flngth

    ngti = 0
    datamode = stringAttribute(set1,"DATAMODE")
    if (datamode /= "IMAGING") then
      ! .and. datamode /= "TIMING" .and. datamode /= "BURST") then
       write (errstr,*) "Mode ( ",datamode," ) not supported"
       call error("badDatamode")
    endif
    write(errstr,*) "Datamode: ",Datamode
    call message(AppMsg,Verbose)
    filterid = stringAttribute(set1,"FILTER")
    if (attinstr == "EPN") then
      if (index(filterid,"OPEN") > 0 .or. &
          index(filterid,"NONE") > 0) then
        nfilter=0 ! 12
      else if (index(filterid,"CLOSED") > 0 .or. &
               index(filterid,"Closed") > 0 .or. &
               index(filterid,"closed") > 0) then
        nfilter=1 ! 13 
      else if (index(filterid,"THIN1") > 0 .or. &
               index(filterid,"Thin1") > 0 .or. &
               index(filterid,"thin1") > 0) then
        nfilter=2 ! 14 
      else if (index(filterid,"THIN2") > 0 .or. &
               index(filterid,"Thin2") > 0 .or. &
               index(filterid,"thin2") > 0) then
        nfilter=3 ! 15 
      else if (index(filterid,"MEDIUM") > 0 .or. &
               index(filterid,"Medium") > 0 .or. &
               index(filterid,"medium") > 0 ) then
        nfilter=4 ! 16 
      else if (index(filterid,"THICK") > 0 .or. &
               index(filterid,"Thick") > 0 .or. &
               index(filterid,"thick") > 0) then
        nfilter=5 ! 17 
      else 
        nfilter=-1
      end if 
    end if 
    if (attinstr == "EMOS1" .or. attinstr == "EMOS2") then
      if (index(filterid,"OPEN") > 0 .or. &
          index(filterid,"NONE") > 0) then
        nfilter=0
      else if (index(filterid,"CLOSED") > 0 .or. &
               index(filterid,"Closed") > 0 .or. &
               index(filterid,"closed") > 0) then
        nfilter=1 
      else if (index(filterid,"THIN1") > 0 .or. &
               index(filterid,"Thin1") > 0 .or. &
               index(filterid,"thin1") > 0 ) then
        nfilter=2 
      else if (index(filterid,"THIN2") > 0 .or. &
               index(filterid,"Thin2") > 0 .or. &
               index(filterid,"thin2") > 0) then
        nfilter=3 
      else if (index(filterid,"MEDIUM") > 0 .or. &
               index(filterid,"Medium") > 0 .or. &
               index(filterid,"medium") > 0) then
        nfilter=4 
      else if (index(filterid,"THICK") > 0 .or. &
               index(filterid,"Thick") > 0 .or. &
               index(filterid,"thick") > 0) then
        nfilter=5 
      else 
        nfilter=-1
      end if 
    end if 
    if (nfilter < 0) then 
      call warning("NoFilt","No FILTER attribute found")
      nfilter=0
    end if
    if (hasAttribute(set1,"SUBMODE")) then  
      submode = stringAttribute(set1,"SUBMODE")
    else
      ! call warning("NoSubMode","No SUBMODE attribute found")
      submode = "PRIME_FULL_WINDOW"
    end if 
    if (attinstr == "EPN") then
      if (index(submode,"PRIME_FULL_WINDOW")  > 0 .or. &
          index(submode,"PrimeFullWindow")  > 0 ) then
        nsubmode=PRIME_FULL_WINDOW
      else if (index(submode,"FULL_WINDOW_EXT") > 0 .or. &
               index(submode,"FullWindowExt") > 0)then
        nsubmode=PRIME_FULL_WINDOW_EXTENDED
      else if (index(submode,"PRIME_LARGE_WINDOW") > 0 .or. &
               index(submode,"PrimeLargeWindow") > 0) then
        nsubmode=PRIME_LARGE_WINDOW
      else if (index(submode,"PRIME_SMALL_WINDOW") > 0 .or. &
               index(submode,"PrimeSmallWindow") > 0) then
        nsubmode=PRIME_SMALL_WINDOW
      else if (index(submode,"PRIME_FULL_OFFSET")  > 0 .or. &
               index(submode,"PrimeFullOffset")  > 0) then
        nsubmode=PRIME_FULL_OFFSET
      else if (index(submode,"PRIME_LARGE_OFFSET") > 0 .or. &
               index(submode,"PrimeLargeOffset") > 0) then
        nsubmode=PRIME_LARGE_OFFSET
      else if (index(submode,"OFFSET") > 0 .or. &
               index(submode,"Offset") > 0 ) then
        nsubmode=OFFSET
      else
        nsubmode=-1
      end if
    end if
    if (attinstr == "EMOS1" .or. attinstr == "EMOS2") nsubmode=-1
      if (index(submode,"PRIME_FULL_WINDOW")  > 0 .or. &
          index(submode,"PrimeFullWindow")  > 0 ) then
        nsubmode=PRIME_FULL_WINDOW
      else if (index(submode,"PRIME_PARTIAL_RFS") > 0 .or. &
               index(submode,"PrimePartialRFS") > 0)then
        nsubmode=PRIME_PARTIAL_RFS
      else if (index(submode,"PRIME_PARTIAL_W2") > 0 .or. &
               index(submode,"PrimePartialW2") > 0) then
        nsubmode=PRIME_PARTIAL_W2
      else if (index(submode,"PRIME_PARTIAL_W3") > 0 .or. &
               index(submode,"PrimePartialW3") > 0) then
        nsubmode=PRIME_PARTIAL_W3
      else if (index(submode,"PRIME_PARTIAL_W4") > 0 .or. &
               index(submode,"PrimePartialW4") > 0) then
        nsubmode=PRIME_PARTIAL_W4
      else if (index(submode,"PRIME_PARTIAL_W5") > 0 .or. &
               index(submode,"PrimePartialW5") > 0) then
        nsubmode=PRIME_PARTIAL_W5
      else if (index(submode,"PRIME_PARTIAL_W6") > 0 .or. &
               index(submode,"PrimePartialW6") > 0) then
        nsubmode=PRIME_PARTIAL_W6
      else
        nsubmode=-1
      end if

    if (nsubmode == -1) then
      ! call warning ("NoSubMode","No valid SUBMODE attribute found")
      if (attinstr == "EPN") then 
        nsubmode = 0
      else
        nsubmode = 0
      end if
    end if  

    shape => dimensions(arr0)
    naxis1 = shape(0)-1
    naxis2 = shape(1)-1

    if (useImageCS) then

       if (hasAttribute(set1,"CTYPE1")) then
         binType=stringAttribute(set1,"CTYPE1")
       else
         binType=stringAttribute(arr0,"CTYPE1")
       endif
       if (useDetector) then
         haswcs2=.false.
         if (trim(binType) /= "DETX" .and. trim(binType) /= "DETY") then
           if (hasAttribute(set1,"CTYPE1L")) then
             binType=stringAttribute(set1,"CTYPE1L")
             haswcs2=.true.
           else
             binType=stringAttribute(arr0,"CTYPE1L")
             haswcs2=.true.
           endif
           if (trim(binType) /= "DETX" .and. trim(binType) /= "DETY") then  
              call error("FileMismatch", &
                         "withdetcoords=yes, Coord. types DETX,DETY expected")
           endif  
         endif          
       else 
         if (trim(binType) /= "RA---TAN" .and. trim(binType) /= "DEC--TAN") then
           call error("FileMismatch", &
                      "withdetcoords=no, Coord. types RA---TAN,DEC--TAN expected")  
         endif  
       endif

       if (useDetector .and. haswcs2) then
!         read secondary WCS header containg DETX,DETY system: 
         if (hasAttribute(set1,"CDELT1L")) then
           detxpx2mm=real32Attribute(set1,"CDELT1L")*flngth * (pi/180.0)
         else
           detxpx2mm=real32Attribute(arr0,"CDELT1L")*flngth * (pi/180.0)
         end if     
         if (hasAttribute(set1,"CRPIX1L")) then 
         detxcrpx=real32Attribute(set1,"CRPIX1L")
         else
           detxcrpx=real32Attribute(arr0,"CRPIX1L")
         end if
         if (hasAttribute(set1,"CRPIX2L")) then
           detycrpx=real32Attribute(set1,"CRPIX2L")
         else
           detycrpx=real32Attribute(arr0,"CRPIX2L")
         end if
         if (hasAttribute(set1,"CRVAL1L")) then  
           detxcrvl=real32Attribute(set1,"CRVAL1L")
         else
           detxcrvl=real32Attribute(arr0,"CRVAL1L")
         end if
         if (hasAttribute(set1,"CRVAL2L")) then
           detycrvl=real32Attribute(set1,"CRVAL2L")
         else
           detycrvl=real32Attribute(arr0,"CRVAL2L")
         end if
         if (hasAttribute(set1,"CDELT2L")) then
           detypx2mm=real32Attribute(set1,"CDELT2L")*flngth * (pi/180.0)
         else
           detypx2mm=real32Attribute(arr0,"CDELT2L")*flngth * (pi/180.0)
         end if
       else
!        read primary WCS header        
         if (hasAttribute(set1,"CDELT1")) then
           detxpx2mm=real32Attribute(set1,"CDELT1")*flngth * (pi/180.0)
         else
           detxpx2mm=real32Attribute(arr0,"CDELT1")*flngth * (pi/180.0)
         end if     
         if (hasAttribute(set1,"CRPIX1")) then 
         detxcrpx=real32Attribute(set1,"CRPIX1")
         else
           detxcrpx=real32Attribute(arr0,"CRPIX1")
         end if
         if (hasAttribute(set1,"CRPIX2")) then
           detycrpx=real32Attribute(set1,"CRPIX2")
         else
           detycrpx=real32Attribute(arr0,"CRPIX2")
         end if
         if (hasAttribute(set1,"CRVAL1")) then  
           detxcrvl=real32Attribute(set1,"CRVAL1")
         else
           detxcrvl=real32Attribute(arr0,"CRVAL1")
         end if
         if (hasAttribute(set1,"CRVAL2")) then
           detycrvl=real32Attribute(set1,"CRVAL2")
         else
           detycrvl=real32Attribute(arr0,"CRVAL2")
         end if
         if (hasAttribute(set1,"CDELT2")) then
           detypx2mm=real32Attribute(set1,"CDELT2")*flngth * (pi/180.0)
         else
           detypx2mm=real32Attribute(arr0,"CDELT2")*flngth * (pi/180.0)
         end if         
       endif
       if (useDetector) then
         detxcrvl=detxcrvl*flngth*pi/(180.0*72000.0)
         detycrvl=detycrvl*flngth*pi/(180.0*72000.0)
         detxpx2mm=detxpx2mm/72000.0
         detypx2mm=detypx2mm/72000.0
       end if
    else
! this is only for development purposes, not for the user
       detxpx2mm=fac
       detxcrpx=naxis1/2.
       detycrpx=naxis2/2.
       detxcrvl=0.
       detycrvl=0.
       detypx2mm=fac
    end if

!   These four quantities, calculated above, are actually wrong
!    for MOS CCDs 2--7, because the common denominator should be
!    (flngth - Zmm(0)).  Because Zmm(0) is not available until later,
!    we save the variables in *_save and fix them later.
    detxcrvl_save  = detxcrvl
    detycrvl_save  = detycrvl
    detxpx2mm_save = detxpx2mm
    detypx2mm_save = detypx2mm

    ranomrad = detxcrvl/radian
    sdecnom = dsin(detycrvl/radian)
    cdecnom = dcos(detycrvl/radian)
!    skypixsize = abs(detypx2mm) / flngth * (180.0/pi)
    skypixsizex = abs(detxpx2mm) / flngth * (180.0d0/pi)
    skypixsizey = abs(detypx2mm) / flngth * (180.0d0/pi)
   


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Remove user pattern selection (23/01/07 GL)
! Event Patterns for QE
!    if(.not. UserPatterns) then
!      ngrade = 0
!      grade(1) = 0
!    else
!       wstr = stringParameter("pattern")
!       lw = len_trim(wstr)
!       ngrade=0
!       i=1
!       do while (i <= lw)
!          junk=i
!          do j=junk,junk
!             if(wstr(j:j) == ' ') exit
!             ngrade=ngrade+1
!             read(wstr(j:min((j+1), lw)),*)grade(ngrade)
!             i=i+1
!          end do
!          i=i+1
!       end do
!       if (attinstr=="EPN") maxgrad=1
!       if (attinstr(1:4)=="EMOS") maxgrad=31
!       ntgrade=0
!       found=.false.
!       do i=0,maxgrad
!         ! GL 17/08/06:
!         found=.false.
!         do j=1,ngrade
!           if (grade(j)==i) found=.true.
!         end do
!         if (.not.found) then
!           ntgrade=ntgrade+1
!           tgrade(ntgrade)=i  
!         end if
!       end do
!       write(*,*) "ngrade,ntgrade,maxgrad",ngrade,ntgrade,maxgrad
!       write(*,*) "grade",grade(0:3)
!      write(*,*) "tgrade",tgrade(0:3)
!    end if
!    write(errstr,*) 'Number of Grades: ',ngrade
!   call message(AppMsg,VERBOSE)
!    write(errstr,*) 'Grades: ',(grade(i),i=1,ngrade)
!    call message(AppMsg,VERBOSE)
!    if(ngrade == 0) then
!      write(errstr,*) 'No grade selection. QE will be set to 1.'
!      call message(AppMsg,VERBOSE)
!    end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!    if (attinstr == "EPN" .and. ngti_ext /= EPN_CCD_LAST) call warning&
!         ("NumGTI", "Number of GTI extensions /= number of chips") 
!    if (attinstr(1:4) == "EMOS" .and. ngti_ext /= EMOS_CCD_LAST) call warning&
!         ("NumGTI", "Number of GTI extensions /= number of chips") 

    if (attinstr == "EPN") then
      ngti_ext = EPN_CCD_LAST
    else
      ngti_ext = EMOS_CCD_LAST
    end if 

    relgti=.false.
    ngti=0
        

    if (useDSS) then

    allocate (tempgti%start(0:0)) 
    allocate (tempgti%stop(0:0))  
    dblock = block(arr0)
    dss = dataSubspace(dblock)
    ncomp = dssComponents(dss)
    acomp=-1
    do icomp = 0, ncomp-1
      !dssc(icomp) = dssComponent(dss,icomp)
      dssc = dssComponent(dss,icomp)
      !nfilt = dssFilters(dssc(icomp))
      nfilt = dssFilters(dssc)
      iccd0=-1
      iccd1=99  
      iccd=0
      irange=0
      itfilt=0  
      ntempgti=0  
      nmask=0
      flagmax(icomp)=0
      flagmin(icomp)=0

      deallocate (tempgti%start)
      deallocate (tempgti%stop)
      allocate (tempgti%start(0:0)) 
      allocate (tempgti%stop(0:0))        

      nflagranges=0 
      do ifilt = 0,nfilt-1
        !dssf(ifilt,icomp) = dssFilter(dssc(icomp),ifilt)
        dssf = dssFilter(dssc,ifilt) 
        !ftype = dssFiltertype(dssf(ifilt,icomp))
        ftype = dssFiltertype(dssf)
!       read bitmask filters for "FLAG" column
        if (ftype == bitmaskfiltert) then
           !fname = trim(dssFilterName(dssf(ifilt,icomp)))
           fname = trim(dssFilterName(dssf))
           !call dssFilterMask(dssf(ifilt,icomp), onBits, offBits)
           call dssFilterMask(dssf, onBits, offBits)
           flagmask(icomp,nmask) = offBits(0)
!            write(*,*) "icomp,ifilt,on,off;", icomp,ifilt,onBits(0), offBits(0)
!           write(*,*)  "size of onBits:", size(onBits) 
           nmask=nmask+1
        endif

        if (ftype == rangefiltert) then
          !fname = trim(dssFilterName(dssf(ifilt,icomp)))
          fname = trim(dssFilterName(dssf))
          !call dssFilterRangesSize(dssf(ifilt,icomp), nSize )
          call dssFilterRangesSize(dssf, nSize )	 
 	      allocate( frange(0:nSize-1)  )
          !frange = dssRanges(dssf(ifilt,icomp))
          frange = dssRanges(dssf) 
          if (fname == "CCDNR") then
            do i=0,size(frange)-1
!              lowB=frange(i)%lowerB
!              lowBval=lowerB%value
!              if (frange(i)%lowerB /= Undef) iccd0(iccd) = int(frange(i)%lower)
!              if (frange(i)%upperB /= Undef) iccd1(iccd) = int(frange(i)%upper)
               lbound = frange(i)%lower
               ubound = frange(i)%upper 
               if (lbound%type /= UNDEFINED) iccd0(iccd) = int(lbound%value)
               if (ubound%type /= UNDEFINED) iccd1(iccd) = int(ubound%value)
               iccd=iccd+1
            end do
          end if

!          read ranges for "FLAG" selections          
          if (fname == "FLAG") then
             nflagranges=nflagranges+1
             do i=0,size(frange)-1
              lbound = frange(i)%lower
              ubound = frange(i)%upper 
              if (lbound%type /= UNDEFINED) then
                flagmintemp=lbound%value
              else
                flagmintemp=0
              end if
              if (ubound%type /= UNDEFINED) then
                flagmaxtemp=ubound%value
              else
                flagmaxtemp=2147483647
              end if
              flagmin(icomp)=min(flagmin(icomp),flagmintemp)
              flagmax(icomp)=max(flagmax(icomp),flagmaxtemp)
              irange=irange+1                  
             end do
          endif

!          read ranges for "TIME" selections
          if (fname == "TIME") then
            do i=0,size(frange)-1
              lbound = frange(i)%lower
              ubound = frange(i)%upper 
              if (lbound%type /= UNDEFINED) then
                fstart(irange)=lbound%value
              else
                fstart(irange)=0d0
              end if
              if (ubound%type /= UNDEFINED) then
                fstop(irange)=ubound%value
              else
                fstop(irange)=1d30
              end if
              irange=irange+1
              if (irange > gtimax) call error("ArrayOutOfRange", &
                 "DSS contains more than 10000 time intervals")
            end do 
            itfilt=itfilt+1
          end if
          deallocate(frange)
        end if 
        if (ftype == gtifiltert) then
          !fname = trim(dssFilterName(dssf(ifilt,icomp)))
          fname = trim(dssFilterName(dssf))
            !gtiblock = dssFilterBlock(dssf(ifilt,icomp))
            gtiblock = dssFilterBlock(dssf)              
            iblock = index(gtiblock,":STDGTI")            
            if (iblock > 0) gtiblock = trim(gtiblock(iblock+1:))
            if (hasBlock(set1,gtiblock)) then
              tabgti = table(set1,gtiblock)
              ntgti = numberOfRows(tabgti) 
              colgtista = column(tabgti,"START", Read)    
              gstart => real64Data(colgtista)
              colgtisto = column(tabgti,"STOP", Read)    
              gstop => real64Data(colgtisto)
              do i=0,ntgti-1
                fstart(irange)=gstart(i)
                fstop(irange)=gstop(i)
                irange = irange+1
                if (irange > gtimax) call error("ArrayOutOfRange", &
                   "DSS contains more than 10000 time intervals")
              end do
            end if
            itfilt=itfilt+1
            call release (colgtista)
            call release (colgtisto)
            call release (tabgti)
        end if  
      end do ! ifilt

       do i=1,ngti_ext
         mgti=.true.
         do j=0,iccd-1
           ic0 = iccd0(j)
           ic1 = iccd1(j)
           !if (.not. (i>ic0.and.i<ic1 .or. i==ic0.and.i==ic1)) merge=.false.
           !if ((i<=ic0.or.i>=ic1).and.(i/=ic0.or.i/=ic1)) merge=.false.
           if ((i<=ic0.or.i>=ic1).and.(i/=ic0.or.i/=ic1)) mgti=.false.
         end do
         if (mgti) then
            !combine filter gtis with &&              
           itfilt1=itfilt+ntempgti-1
           call merge_gti(tempgti, ntempgti, fstart, fstop, irange, itfilt, itfilt1)
            !combine component gtis with ||
           itfilt=ntempgti
           itfilt1=0
           do j=0,ntempgti-1
             tempstart(j)=tempgti%start(j)
             tempstop(j)=tempgti%stop(j)
           end do
           irange=ntempgti
           call merge_gti(gti(i), ngti(i), tempstart, tempstop, irange, itfilt, itfilt1)
           acomp(i)=icomp  
         end if   
       end do        
         
    end do ! icomp
    if (nflagranges == 0) then
      flagmax=2147483647
    endif    


    
    ntest=0
    do i=1,ngti_ext
      ntest=ntest+ngti(i)
    end do 
    if (ntest==0) call error ("NoGTI", &
                  "No GTI or TIME filter in DSS")

    else ! now part for usedss=.false.

! GTI extensions with the names STDGTI, STDGTI1 .. STDGTI9, 
! STDGTI01 .. STDGTI12 are accepted
    hasgti=.false.
    name = "STDGTI"
    if (hasBlock(set1,name)) then
      hasgti=.true.
      tabi(1) = table(set1,name)  
      ngti1 = numberOfRows(tabi(1))
      colstart(1) = column(tabi(1),"START", Read)    
      gti(1)%start => real64Data(colstart(1))
      colstop(1) = column(tabi(1),"STOP", Read)    
      gti(1)%stop => real64Data(colstop(1))
      relgti(1)=.true.
!      ngti_ext = 1
    else
      name = "GTI" ! to work with Xselect output
      if (hasBlock(set1,name)) then
         hasgti=.true. 
         tabi(1) = table(set1,name)  
         ngti1 = numberOfRows(tabi(1))
         colstart(1) = column(tabi(1),"START", Read)    
         gti(1)%start => real64Data(colstart(1))
         colstop(1) = column(tabi(1),"STOP", Read)    
         gti(1)%stop => real64Data(colstop(1))
         relgti(1)=.true. 
!         ngti_ext = 1
      end if
    end if
    hasgtin=.false.
    do i = 1,nextmax
      write(name,"(a,i2.2)") "STDGTI", i  
      if (hasBlock(set1,name)) then
        hasgti=.true.
        hasgtin=.true. 
        if (relgti(i)) then
          call release(colstart(i))
          call release(colstop(i))
          call release(tabi(i))
        end if
        tabi(i) = table(set1,name)
        ngti(i) = numberOfRows(tabi(i))
        colstart(i) = column(tabi(i),"START", Read)    
        gti(i)%start => real64Data(colstart(i))
        colstop(i) = column(tabi(i),"STOP", Read)    
        gti(i)%stop => real64Data(colstop(i))
        relgti(i)=.true.
!        ngti_ext = i
      end if
      if (i < 11) then
        write(name,"(a,i1)") "STDGTI", i   
        if (hasBlock(set1,name)) then
          hasgti=.true.
          hasgtin=.true. 
          tabi(i) = table(set1,name)
          ngti(i) = numberOfRows(tabi(i))
          colstart(i) = column(tabi(i),"START", Read)    
          gti(i)%start => real64Data(colstart(i))
          colstop(i) = column(tabi(i),"STOP", Read)    
          gti(i)%stop => real64Data(colstop(i))
          relgti(i)=.true.
!          ngti_ext = i
        end if
      end if
    end do

    if (.not. hasgti) then

     call warning ("noGTI","No GTI extension found in input image") 

! start and stop are set to moderately meaningful values in case 
! no GTI extensions are present. 
    if (hasAttribute(set1,"EXPOSURE")) then
      expos = real32Attribute(set1,"EXPOSURE")
! fix to work with early data. Alexis
!     if(expos .gt. 1.e5) expos=1.
!!!      ngti(1) = 1
      tstart(0) = time(0)
      tstop(0) = time(0) + expos
      do i=1,ngti_ext 
        gti(i)%start => tstart
        gti(i)%stop => tstop
        ngti(i)=1    
      end do
      ngti1=1
!      ngti_ext = 0
    else if (hasAttribute(arr0,"EXPOSURE")) then
      expos = real32Attribute(arr0,"EXPOSURE")
! fix to work with early data
!     if(expos .gt. 1.e5) expos=1.
!!!      ngti(1) = 1
      tstart(0) = time(0)
      tstop(0) = time(0) + expos
      do i=1,ngti_ext
        gti(i)%start => tstart
        gti(i)%stop => tstop
        ngti(i)=1      
      end do
      ngti1=1
!      ngti_ext = 0
    else
      call error ("onGTI/noexposure", &
                  "No GTI and no exposure available")
    end if 

    end if

    if (.not.hasgtin) then
      ngti(1)=ngti1
      do i=2, ngti_ext
        allocate (gti(i)%start(0:ngti(1)-1))
        allocate (gti(i)%stop(0:ngti(1)-1))
        gti(i)%start = gti(1)%start
        gti(i)%stop = gti(1)%stop
        ngti(i)=ngti(1) 
       end do
     end if
      
!    ngti_ext0 = ngti_ext 
!    if (ngti_ext > 0) then
!      ngti_ext0 = ngti_ext
!    else
!      ngti_ext0 = 0
!      ngti_ext = 1
!      ngti(1) = 1
!    end if

    end if ! usedss

    if(attinstr .ne. 'EPN') then
       nccd0=EMOS_CCD_FIRST
       nccd=EMOS_CCD_LAST
    else
       nccd0=EPN_CCD_FIRST
       nccd=EPN_CCD_LAST
    end if

! prepare to read the exposure extension and bad pixel stuff
    set3=dataset(stringParameter("eventset"),Read)
    bbname = "BADPIX"
    ocname = "OFFSET"

! create exposure images
    do i=0,nband-1
      setO(i)=dataset(stringParameter("expimageset",i),Create)
      arr(i)=addArray(setO(i),"PRIMARY",Real32,shape)
      earr(i)%exp => real32Array2Data(arr(i))          
      earr(i)%exp = 0.
    end do
    yextra=0.

! filterId needs to be read from input image
    if (attinstr.eq.'EPN') then
       call CAL_SetState(instrument=EPN, &
            filterId=nfilter, ccdModeId=nsubmode, randomize=rndmz)

       call OAL_SetState(instrument=EPN)
       imod = 5
!       nsplit=max(4,ceiling(0.5*MM_PX_X/detxpx2mm),ceiling(0.5*MM_PX_Y/detypx2mm))
!      detxpx2mm is negative for sky images.
       nsplit=max(4,ceiling(0.5*MM_PX_X/abs(detxpx2mm)),&
                    ceiling(0.5*MM_PX_Y/abs(detypx2mm))  )
       xhsz_mm= 0.5* MM_PX_X
       yhsz_mm= 0.5* MM_PX_Y
!       if(ipy .eq. EPN_MAX_Y) yhsz_mm= 0.5* MM_PX_LY
       yextra(EPN_MAX_Y) = 0.5* MM_PX_LY - yhsz_mm
       xshift=0.
       yshift=0.
    end if

    if (attinstr.eq.'EMOS1') then
       call CAL_SetState(instrument=EMOS1, &
            filterId=nfilter,ccdNodeId=0,ccdModeId=nsubmode, randomize&
            & =.false.)
       call OAL_SetState(instrument=EMOS1)
       imod = 20
!       nsplit=max(1,ceiling(0.5*MOS_PX/detxpx2mm),ceiling(0.5*MOS_PX/detypx2mm))
!      detxpx2mm is negative for sky images.
       nsplit=max(1,ceiling(0.5*MOS_PX/abs(detxpx2mm)),&
                    ceiling(0.5*MOS_PX/abs(detypx2mm))  )
       xhsz_mm= 0.5* MOS_PX
       yhsz_mm= 0.5* MOS_PX
       xshift=0.
       yshift=0.
    end if
       
    if (attinstr.eq.'EMOS2') then
       call CAL_SetState(instrument=EMOS2, &
            filterId=nfilter,ccdNodeId=0,ccdModeId=nsubmode, randomize&
            & =.false.)
       call OAL_SetState(instrument=EMOS2)
       imod = 20
!       nsplit=max(1,ceiling(0.5*MOS_PX/detxpx2mm),ceiling(0.5*MOS_PX/detypx2mm))
!      detxpx2mm is negative for sky images.
       nsplit=max(1,ceiling(0.5*MOS_PX/abs(detxpx2mm)),&
                    ceiling(0.5*MOS_PX/abs(detypx2mm))  )
       xhsz_mm= 0.5* MOS_PX
       yhsz_mm= 0.5* MOS_PX
       xshift=0.
       yshift=0.
    end if


!   pixoff is now obsolete  
!    pixoff=2.0*xhsz_mm/abs(detxpx2mm)*(0.5/float(nsplit)-0.5)


   ! Calculate the position of the optical axis in the CamCoord2 frame
   ! RDS
    
    call CAL_setState(ccdChipId= &
                          int(CAL_getMiscellaneousDataValue('OPTICS_CCD')))
    opticsx(0) = int(CAL_getMiscellaneousDataValue('OPTICS_X'))
    opticsy(0) = int(CAL_getMiscellaneousDataValue('OPTICS_Y'))
                                                  
!    write(*,*) " opticsx(0), opticsy(0)", opticsx(0), opticsy(0)
                              
    call CAL_chipCoordToCamCoord1(opticsx, opticsy, &
                                  opticsx_cc, opticsy_cc, opticsz_cc);
    call CAL_camCoord1ToCamCoord2(opticsx_cc, opticsy_cc, opticsz_cc, &
                                  opticsx_cc, opticsy_cc, opticsz_cc);
   ! End of RDS

!    write(*,*) " opticsx_cc(0), opticsy_cc(0)", opticsx_cc(0), opticsy_cc(0)


    pixfac=2.0*xhsz_mm/abs(detxpx2mm)/float(nsplit)
    fhx=xhsz_mm/(abs(detxpx2mm)*nsplit)
    fhy=yhsz_mm/(abs(detypx2mm)*nsplit)

    fh2=4.0*fhx*fhy

! we were advised to use medium quality 
    call CAL_setState(accuracyLevel = ACCURACY_MEDIUM)

!    if (.not. useDetector) then
! precalculation of the detector dead times using the GTI. For sky
    ! coordinates, however, this has to be done on pointing histogram time
    ! step. 

!       call atthst (ranomrad, sdecnom, cdecnom, skypixsize, time, ra0, de0,&
       call atthst (ranomrad, sdecnom, cdecnom, skypixsizex, skypixsizey, time, ra0, de0,&
            pa0, ncoras, gti, ngti, ngti_ext, nccd, set3,  & ! input 
            xnew, ynew, crnew, srnew, natt, eps, &
            winx0, winy0, windx, windy, nextmax, ootfrac)        
!       print*,xnew(1),ynew(1)
!    end if

! Here we allocate the memory for pixel transformation arrays
    allocate (qalpha(1:natt))
    allocate (alsin(1:natt))
    allocate (alcos(1:natt))
    allocate (hs(1:natt))
    allocate (q45(1:natt))
    do k=1,natt
       qalpha(k) = .true.
    enddo
 

!! 80 is to suppress errors in attcalc 
!    if(.not.fastmode) skypixsize = skypixsize/80. 

    if (fastmode.and.natt.gt.18) then
      fastmode=.false.
!      skypixsize = skypixsize/80.
      call warning("NoFastPixelization", &
      "Too many attitude bins; fast pixelization turned off")
    end if

    if (fastmode) then
      allocate (exp0(1:natt,0:naxis1-1,0:naxis2-1)) 
      exp0(1:natt,0:naxis1-1,0:naxis2-1)=0.0   
      allocate (exp1(0:nband-1,1:natt,0:naxis1-1,0:naxis2-1)) 
      exp1(0:nband-1,1:natt,0:naxis1-1,0:naxis2-1)=0.0   
    end if

    if (useDetector) then
      do iccd=1,nccd
        do k=2,natt
          fexpext(iccd,1)=fexpext(iccd,1)+fexpext(iccd,k)
        end do
      end do
      natt=1 
    end if 

    do i = 1, nextmax
      if (relgti(i)) then
        call release(colstart(i)) 
        call release(colstop(i))
        call release(tabi(i))
      end if
    end do
    call release(colrao)
    call release(coldeo)
    call release(colpa)
    call release(coltime)
    call release(tabcoras)
   
!    call CAL_piToeV(amplo,emin)
!    call CAL_piToeV(amphi,emax)
!    e(0) = (emin(0)+emax(0))/2.0
    
!     do i=0,nband-1
!       e(i) = (amplo(i)+amphi(i))/2.0
!       if (withVig) then
!         effarea_cen(i) = CAL_getEffectiveArea(e(i),0d0,0d0) * &
!                          CAL_getFilterTransmission(e(i),0d0,0d0)
!       else
!         effarea_cen(i) = 1.0
!       end if
!     end do
!    Rewritten for speed-up as:
     if (withVig) then
       do i=0,nband-1
         e(i) = (amplo(i)+amphi(i))/2.0
         effarea_cen(i) = CAL_getEffectiveArea(e(i),0d0,0d0) * &
                          CAL_getFilterTransmission(e(i),0d0,0d0)
       end do
     else
       do i=0,nband-1
         effarea_cen(i) = 1.0
       end do
     end if

! effective area at optical axis

    !------------------------------------------!
    !   get  CC2->SC Cmatrix and angles        !
    !------------------------------------------!
    do i=nextmax,1,-1
      if (ngti(i)>0) ctime=gti(i)%start(0)
    end do 
    call CAL_getCamCoord2ToSacCoord(ctime, Cmatrix)
    bmtrx10 = Cmatrix(1,0)
    bmtrx11 = Cmatrix(1,1)
    bmtrx20 = Cmatrix(2,0)
    bmtrx21 = Cmatrix(2,1)

!    write (*,*) Cmatrix(1,0),Cmatrix(1,1),Cmatrix(1,2),Cmatrix(2,0),Cmatrix(2,1)
!*     write (*,*) Cmatrix(2,0),Cmatrix(2,1),Cmatrix(2,2)
    call CAL_releaseMemory(Cmatrix)

! read EPN OFFSETS extension:
    hasoffset0=.false.
    if (attinstr .eq. 'EPN' .and. useoffset) then
      if (hasBlock(set3,"OFFSETS")) then
        taboffs = table(set3,"OFFSETS")
        nofs = numberOfRows(taboffs) 
        if (nofs > 0) then
            cox = column(taboffs,"RAWX", Read)    
            bbox => int16Data(cox)
            cofs = column(taboffs,"OFFSETX", Read)
            bbofs => int16Data(cofs)
            cnr  = column(taboffs,"CCDNR", Read)
            bbnr  => int8Data(cnr)
            hasoffset0=.true.
            write(errstr,*) 'read offsets for ',nofs, ' columns'
            call message(AppMsg,VERBOSE)
        endif               
      endif
    endif



! read MOS OFFSET extension
    if ((attinstr .eq. 'EMOS1' .or. attinstr .eq. 'EMOS2')  .and. useoffset) then
      if (hasBlock(set3,"OFFSETS")) then
        taboffs = table(set3,"OFFSETS")
        nofs = numberOfRows(taboffs) 
        if (nofs > 0) then
            cox = column(taboffs,"RAWX", Read)    
            bbox => int16Data(cox)
            coffx = column(taboffs,"OFFSETX", Read)
            bboffx => int16Data(coffx)
            coffy = column(taboffs,"OFFSETY", Read)
            bboffy => int16Data(coffy)
            cnr  = column(taboffs,"CCDNR", Read)
            bbnr  => int8Data(cnr)
            hasoffset0=.true.
            write(errstr,*) 'read offsets for ',nofs, ' columns'
            call message(AppMsg,VERBOSE)
        endif
      else 
        call warning("NoOffset","Extension OFFSET not found")               
      endif
    endif


!   Certain quantities are constant in the ccd loop below.
!   We thus assign them values once and for all.
    if (attinstr .eq. 'EPN') then
      npx0 = EPN_MIN_X
      npx1 = EPN_MAX_X
      npy0 = EPN_MIN_Y
      npy1 = EPN_MAX_Y
    else
      npx0 = EMOS_MIN_X
      npx1 = EMOS_MAX_X
      npy0 = EMOS_MIN_Y
      npy1 = EMOS_MAX_Y
    end if
    do ipy=npy0,npy1 
      rawY(ipy)=ipy
    end do 

!-----------------------------------------
    xxname1 = "DLIMAP"              !-- JR
!-----------------------------------------

! RDS - get MISClengtlengthh data values once
    rfac = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X') * 0.05d0 &
           / CAL_getMiscellaneousDataValue('PLATE_SCALE_X') &
		* 180.0d0 * 72000.0d0 / flngth / pi

    rfac2x = CAL_getMiscellaneousDataValue('PLATE_SCALE_X')&
                          /CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')&
                          *20.0d0 
    rfac2y = CAL_getMiscellaneousDataValue('PLATE_SCALE_Y')&
                          /CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y')&
                          *20.0d0 

!loop over ccds.
    close=.false.
    closeos=.false.
    do iccd=nccd0, nccd
!!    do iccd=7,7 ! for quick test
!-----------------------------------------
!-----------------------------------------
! read EPN DLMAP** extension:       !-- JR
!-----------------------------------------
       write(name1,"(i2.2)") iccd
       name1=xxname1//name1(1:2)
       hasdliset0=.false.
!---
    if (attinstr .eq. 'EPN') then  !! Discuss a new parameter?
      if (hasBlock(set3,name1)) then
        tabdliexpco = table(set3,name1)
        ndli = numberOfRows(tabdliexpco)
        if (ndli > 0 .and. hasColumn(tabdliexpco,"DLIEXPCO")) then
           dliexpco1 = column(tabdliexpco,"DLIEXPCO", Read)
           bdliexpco1 => real32Data(dliexpco1)
           hasdliset0=.true.
           write(errstr,*) '*** read dliexpco column for correction ***',ndli, ' columns'
           call message(AppMsg,VERBOSE)
        endif
      endif
    endif
!-----------------------------------------
!-----------------------------------------
      if (ngti(iccd) > 0) then 
       write(errstr,*) 'doing ',attinstr,' ccd chip #',iccd
       call message(AppMsg,VERBOSE)
       call CAL_setState(ccdChipId=iccd)
       call CAL_getFOVmap(map)
!      Constant expressions moved outside of ccd loop:
!       npx0=EPN_MIN_X
!       npx1=EPN_MAX_X
!       npy0=EPN_MIN_Y
!       npy1=EPN_MAX_Y
       mpx0=max(winx0(iccd),EPN_MIN_X)
       mpx1=min(mpx0+windx(iccd)-1,EPN_MAX_X)
       mpy0=max(winy0(iccd),EPN_MIN_Y)
       mpy1=min(mpy0+windy(iccd)-1,EPN_MAX_Y)
       if(attinstr .ne. 'EPN') then
!      Constant expressions moved outside of ccd loop:
!         npx0=EMOS_MIN_X
!         npx1=EMOS_MAX_X
!         npy0=EMOS_MIN_Y
!         npy1=EMOS_MAX_Y
         mpx0=max(winx0(iccd),EMOS_MIN_X)
         mpx1=min(mpx0+windx(iccd)-1,EMOS_MAX_X)
         mpy0=max(winy0(iccd),EMOS_MIN_Y)
         mpy1=min(mpy0+windy(iccd)-1,EMOS_MAX_Y)
       end if


! Get margins defined by flag selections:
      
!      write(*,*) "flagmin:",flagmin
!      write(*,*) "flagmax:",flagmax
!      write(*,*) "flagmask:",flagmask(0:11,0:1)
!      write(*,*) "margins:",flagmargins(1:3),flagmargins2(1:3),bordermargin,offsetmargin

      call get_margins(attinstr,iccd-1,flagmin,flagmax,flagmask,nmask,&
                       flagmargins,flagmargins2,bordermargin,offsetmargin)

! Read Bad pixel extension
!       if(nbb > 0 .and. iccd > 1) then
        if (close) then 
! clean previous extension
          call release (crawx)
          call release (crawy)
          call release (cye)
          call release (tabcoras)
       end if
        if (closeos) then 
! clean previous extension
          call release (cox)
          call release (cofs)
          call release (taboffs)
          closeos=.false.
       end if
       write(name,"(i2.2)") iccd  
       name=bbname//name(1:2)
       if (hasBlock(set3,name)) then
          tabcoras = table(set3,name)
          nbb = numberOfRows(tabcoras)
          crawx = column(tabcoras,"RAWX", Read)    
          bbrawx => int16Data(crawx)
          crawy = column(tabcoras,"RAWY", Read)
          bbrawy => int16Data(crawy)
          cye = column(tabcoras,"YEXTENT", Read)
          bbye => int16Data(cye)
          cbadtype = column(tabcoras,"TYPE", Read)
          bbadtype => int16Data(cbadtype)
          cbadflag = column(tabcoras,"BADFLAG", Read)
          bbadflag => int16Data(cbadflag)
          close=.true.
       else
          nbb=0
          call warning("NoBadPix","BadPixel extension not found")
       end if  
       allocate(badpixtype(0:nbb-1)) 
       !!! IT 2017/06 initialize badpixtype with zero if compiling with gfortran
       badpixtype(0:nbb-1) = 0

!      translate flags from BADPIX extensions to relevant event flags: 
         do ibb=0,nbb-1 
           if (bbadtype(ibb) == 1 .and. bbadflag(ibb) == 1) badpixtype(ibb)=1
           if (bbadtype(ibb) == 1 .and. bbadflag(ibb) > 1)  badpixtype(ibb)=3 
           if  (bbadtype(ibb) == 3)  badpixtype(ibb)=2 
         enddo  

! get the information about the offset columns
       if (attinstr .eq. 'EPN' .and. useoffset .and. .not. hasoffset0) then
          write(name,"(i2.2)") iccd  
          name=ocname//name(1:2)
          if (hasBlock(set3,name)) then
             taboffs = table(set3,name)
             nofs = numberOfRows(taboffs)
             if (nofs > 0 .and. hasColumn(taboffs,"RAWX")) then
                cox = column(taboffs,"RAWX", Read)    
                bbox => int16Data(cox)
                cofs = column(taboffs,"OFFSET", Read)
                bbofs => int16Data(cofs)
                closeos=.true.
                write(errstr,*) 'read offsets in a number of',nofs
                call message(AppMsg,VERBOSE)
             endif
          else
             closeos=.false.
             call warning("NoOffset","EPN Offset extension not found")
          end if
       end if
               
!      offsetflag constant for each CCD.  Moved from ipx loop below.
       offsetflag=.false.
       if (attinstr == "EPN") then
         if (flagmax(iccd) < 8) then
           offsetflag=.true. 
         else
           do imask=0,nmask-1
             if (btest(flagmask(iccd,imask),3)) offsetflag=.true.
           end do
         endif
       endif
       if (attinstr == "EMOS1" .or. attinstr == "EMOS2") then
         if (flagmax(iccd) < 8192) then
           offsetflag=.true. 
         else
           do imask=0,nmask-1
             if (btest(flagmask(iccd,imask),13)) offsetflag=.true.
           end do
         endif
       endif

       area_save=-1.0  

!       write(*,*)'Offset'

! get bad pixel identificator
       rbbc=get_BadRadius(ibbc)
       ibbc1=rbbc
       irbbc2=rbbc**2


! loop over the pixels, PIXCOORD1
       do ipx=npx0, npx1 
          qofs=.false.
          if (modulo(ipx,imod) == 0) area_save=-1.0

          do ipy=npy0,npy1 
            rawX(ipy)=ipx
!            rawY(ipy)=ipy
!           rawY(:) stay constant through the subroutine; moved outside ccd loop
          end do 

!  removed 15/11/2002:
!          nbb0=nbb+1
!          nbb1=0
!          do ibb=0,nbb-1
!            if (bbrawx(ibb) >= ipx-ibbc1 .and. &
!                bbrawx(ibb) <= ipx+ibbc1) then
!              nbb0=min(nbb0,ibb)
!              nbb1=max(nbb1,ibb)
!            end if
!          end do
!          write(*,*) "nbb0,nbb1:",nbb0,nbb1


!  The offsetflag is constant for each CCD.
!  The following lines moved out of the loop over ipx.
!   offsetflag=.false.
!   if (attinstr == "EPN") then
!     if (flagmax(iccd) < 8) offsetflag=.true. 
!     do imask=0,nmask-1
!        if (btest(flagmask(iccd,imask),3)) offsetflag=.true.
!     end do
!   endif
!   if (attinstr == "EMOS1" .or. attinstr == "EMOS2") then
!     if (flagmax(iccd) < 8192) offsetflag=.true. 
!     do imask=0,nmask-1
!        if (btest(flagmask(iccd,imask),13)) offsetflag=.true.
!     end do
!   endif



!MOS OFFSET columns:
         if((attinstr .eq. 'EMOS1' .or. attinstr .eq. 'EMOS2') .and. hasoffset0 .and. offsetflag) then
                do iofs=0,nofs-1
                   if(ipx == bbox(iofs) .and. bboffx(iofs) > 0 .and. bbnr(iofs) == iccd) then
                      write(errstr,*) 'offset column ',  bbox(iofs),' will be excluded'
                      call message(AppMsg,VERBOSE)
                      qofs =.true.
                   endif
                end do
          endif
 
!write(*,*) "attinst,hasoff,offsetflag,offsetmargin",attinstr,hasoffset0,offsetflag,offsetmargin


! EPN offset columns
          if(attinstr .eq. 'EPN' .and. hasoffset0 .and. offsetflag) then
! use margin from flag selections or badclean parameter, whichever is larger:
             margin=max(ibbc1,offsetmargin)
             do ipxad=ipx-margin,ipx+margin
                do iofs=0,nofs-1
                   if(ipxad == bbox(iofs) .and. bbofs(iofs) > hducut .and. bbnr(iofs) == iccd) then
                      write(errstr,*) 'offset column ',  bbox(iofs),' will be excluded'
                      call message(AppMsg,VERBOSE)
                      qofs =.true.
                   endif
                end do
             end do
          endif
          if(closeos) then
             margin=max(ibbc1,offsetmargin)
             do ipxad=ipx-margin,ipx+margin
                do iofs=0,nofs-1
                   if(ipxad == bbox(iofs) .and. bbofs(iofs) > hducut) then
                      write(errstr,*) 'offset column ',  bbox(iofs),' will be excluded'
                      call message(AppMsg,VERBOSE)
                      qofs =.true.
                   endif
                end do
             end do
          endif

          call CAL_rawXY2mm(rawX(npy0:npy1),rawY(npy0:npy1),Xmm,Ymm,Zmm,2)

!
!         Now that Zmm(0) is available, recalculate the previously saved,
!          and wrong, conversion factors properly.  Note that all elements
!          of Zmm(:) are identical.
          if (attinstr == "EPN") then
            detxpx2mm = detxpx2mm_save * rfac
!            detxpx2mm = detxpx2mm_save &
!		* CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X') * 0.05d0 &
!		/ CAL_getMiscellaneousDataValue('PLATE_SCALE_X') &
!		* 180.0d0 * 72000.0d0 / flngth / pi
            detypx2mm = detypx2mm_save * rfac
!            detypx2mm = detypx2mm_save &
!		* CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y') * 0.05d0 &
!		/ CAL_getMiscellaneousDataValue('PLATE_SCALE_Y') &
!		* 180.0d0 * 72000.0d0 / flngth / pi
            if (useDetector) then
              detxcrvl = detxcrvl_save * rfac
!              detxcrvl = detxcrvl_save &
!		* CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X') * 0.05d0 &
!		/ CAL_getMiscellaneousDataValue('PLATE_SCALE_X') &
!		* 180.0d0 * 72000.0d0 / flngth / pi
              detycrvl = detycrvl_save * rfac
!              detycrvl = detycrvl_save &
!		* CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y') * 0.05d0 &
!		/ CAL_getMiscellaneousDataValue('PLATE_SCALE_Y') &
!		* 180.0d0 * 72000.0d0 / flngth / pi
            end if
          else
            detxpx2mm = detxpx2mm_save * (flngth - Zmm(0)) / flngth
            detypx2mm = detypx2mm_save * (flngth - Zmm(0)) / flngth
            if (useDetector) then
              detxcrvl = detxcrvl_save * (flngth - Zmm(0)) / flngth
              detycrvl = detycrvl_save * (flngth - Zmm(0)) / flngth
            end if
          end if

!! Removed from HUbert version for being time and memory consuming  (2/2/2010 - JR)
!       call CAL_getFOVmap(map)

          call CAL_camCoord2ToTelCoord(Xmm(npy0-1:npy1-1),Ymm(npy0-1:npy1-1),&
                                       Zmm(npy0-1:npy1-1),theta, phi)

          notsaid=.true.

          
          do ipyp=npy0,npy1
             do ipy=ipyp,ipyp ! this is done to accomplish different exits
! check if this pixel is bad or out of FOV
!                if(CAL_isPixelBad(ipx,ipy,yextent,defect,uplinked)) exit
                qbb=.false.
                qofsy=.false.
                if(map(ipx-1,ipy-1) /= 1) qbb=.true.
                if (ipx<mpx0 .or. ipx>mpx1 .or. &
                    ipy<mpy0 .or. ipy>mpy1) qbb=.true. 


! MOS bad offset rows:
! If bboffy(iofs) > 0, interpret bbox(=RAWX) as RAWY
                if((attinstr .eq. 'EMOS1' .or. attinstr .eq. 'EMOS2') .and. hasoffset0) then
                 do iofs=0,nofs-1
                    if(ipy == bbox(iofs) .and. bboffy(iofs) > 0 .and. bbnr(iofs) == iccd) then
                      qofsy =.true.
                    endif
                 end do
                endif

!  ??????        do ibb=nbb0,nbb1
                 do ibb=0,nbb-1
!                 use badpix margin from flag selection or badclean parameter, whichever is larger:
                  if ((badpixtype(ibb) > 0)) then                
                    margin=max(ibbc1,flagmargins(badpixtype(ibb)))
                    margin2=max(irbbc2,flagmargins2(badpixtype(ibb)))
                  else  ! pixels with badpixtype=0 are "intact"
                    margin=0
                    margin2=0
                  endif
                  do ipxad=ipx-margin,ipx+margin
                    if (ipxad==bbrawx(ibb)) then  
                      ipxd2=(ipxad-ipx)**2
                      bbrawy0=bbrawy(ibb)
                      bbrawy1=bbrawy0+bbye(ibb) 
                      do ipyad=ipy-margin,ipy+margin
                        if (ipyad >= bbrawy0 .and. ipyad < bbrawy1) then
                            irpx2=ipxd2+(ipyad-ipy)**2
                            if(irpx2 <= margin2) qbb =.true.
                        end if
                      end do
                    end if
                  end do
                end do
!               if(qbb .or. qofsy) exit
! Borders 
!               use border margin from flag selection or badclean parameter, whichever is larger: 
                                  
                margin=max(ibbc1,bordermargin)
!                write(*,*) "bordermargins:",ibbc1,bordermargin
                do ipxad=ipx-margin,ipx+margin
                   do ipyad=ipy-margin,ipy+margin
                      if(ipxad == npx0-1 .or. ipxad == npx1+1 .or.&
                           ipyad== npy0-1 .or. ipyad == npy1+1 ) qbb =.true.
                   end do
                end do
!               if(qbb) exit
!             if (qsmpl) exit
                              
! transform PIXCOORD1 --> CAMCOORD2
                ! 2 stands for PIXCOORD1 --> CAMCOORD2 transformation

                if (area_save(ipy/imod,0)< 0) then
 
!#########   always effgrad=1. (GL 23/07/2007) 
                  effgrad=1.
!#########  This is obsolete without pattern selection:
!                  if (ngrade==0) then
!                    effgrad=1.
!                  else 
!                    ipym2=ipy+imod/2
!                    effgrad=0.
!                    do igrade=1,ngrade                  
!                      do i=0,nband-1
!                        ! get quantum efficiency 
!                        effgrad(i) = effgrad(i) + &
!                        CAL_getQuantumEfficiency(e(i),ipx,ipym2,grade(igrade))
!                      end do
!                    end do
!                   teffgrad=0.
!                    do igrade=1,ntgrade                  
!                      do i=0,nband-1
!                        ! get quantum efficiency 
!                        teffgrad(i) = teffgrad(i) + &
!                        CAL_getQuantumEfficiency(e(i),ipx,ipym2,tgrade(igrade))
!                      end do
!                    end do
!                    do i=0,nband-1
!!                      write(*,*)  ipx,ipy,ngrade,ntgrade,effgrad(i),teffgrad(i)
!                      effgrad(i)=effgrad(i)/(effgrad(i)+teffgrad(i))
!                    end do
!                  end if

                  dtheta=theta(ipy-1)

                  dtheta=min(theta(ipy-1),999.0d0)  
!!!!!!                  dtheta=min(theta(ipy-1),999.0)  
                  dphi=phi(ipy-1)
                  if (dphi < 0d0) dphi = 2d0*pi + dphi
                  if (withVig) then
                    do i=0,nband-1

                      area(i)=CAL_getEffectiveArea(e(i),dtheta,dphi) * &
                              CAL_getFilterTransmission(e(i),dtheta,dphi) / &
                              effarea_cen(i) * effgrad(i)
                      area_save(ipy/imod,i) = area(i)  
                    end do 
                  else
                    do i=0,nband-1
                      area(i)=effgrad(i)
                      area_save(ipy/imod,i) = area(i)  
                    end do 
                  end if 
                else
                  do i=0,nband-1
                    area(i) = area_save(ipy/imod,i)
                  end do
                end if 
                
                if (qbb .or. qofs .or. qofsy) area(0:nband-1)=0.0
!--------------------------------------------------------------------------
!------------------ DLIEXPCO correction factor (12/02/09 - JR) ------------
                if (dlimode) then
                  if (hasdliset0) then
                    do ia=0,nband-1
                    area(ia) = area(ia) * bdliexpco1(ipx)
                    !write(*,*) "### hello!! ###",ia,iccd,bdliexpco1(ipx)
                    enddo
                  endif
                endif
!--------------------------------------------------------------------------
!--------------------------------------------------------------------------

                if (useDetector) then
                   ! account for the pixel size

!---------------------------------------------------------------------------------------------------
! change back to this version, assuming Cam-coordinates refer to centres of Raw pixels (22/05/07, GL)
! changed in the CAL?  

                    xexpu=(Xmm(ipy-1)+xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
                    xexpl=(Xmm(ipy-1)-xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
                    yexpu=(Ymm(ipy-1)+yhsz_mm - detycrvl)/detypx2mm + detycrpx
                    yexpl=(Ymm(ipy-1)-yhsz_mm - detycrvl)/detypx2mm + detycrpx


!-----------------------------------------------------------------------------------------------
! account for the pixel size
!-----------------------------------------------------------------------------------------------
!                   xexpu=(Xmm(ipy-1) - detxcrvl)/detxpx2mm + detxcrpx
!                   xexpl=(Xmm(ipy-1)-2*xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
!                   yexpu=(Ymm(ipy-1) - detycrvl)/detypx2mm + detycrpx
!                   yexpl=(Ymm(ipy-1)-2*(yhsz_mm+yextra(ipy)) - detycrvl)/detypx2mm + detycrpx
!-----------------------------------------------------------------------------------------------

!                   iexp0=xexpl
!                   iexp1=xexpu
!                   jexp0=yexpl
!                   jexp1=yexpu
!                  The formulae above cast doubles to ints, flooring the values
!                   in the process, which is not what we intend.  Instead,
!                   we want to round [0.5, 1.5) to 1.  nint() does that.
                   iexp0=nint(xexpl)
                   iexp1=nint(xexpu)
                   jexp0=nint(yexpl)
                   jexp1=nint(yexpu)
                   
!*                 if(xexpl .lt. 0) iexp0=iexp0-1
!*                 if(xexpu .lt. 0) iexp1=iexp1-1
!*                 if(yexpl .lt. 0) jexp0=jexp0-1
!*                 if(yexpu .lt. 0) jexp1=jexp1-1
! this is true, but not necessary, since we discard these values later

                   do iexp=iexp0,iexp1
                      do jexp=jexp0,jexp1
!                         if(iexp .ge. 0 .and. iexp .le. naxis1 .and. &
!                            jexp .ge. 0 .and. jexp .le. naxis2) then
!                            fracx=min(xexpu,real(iexp+1))-max(xexpl,real(iexp))
!                            fracy=min(yexpu,real(jexp+1))-max(yexpl,real(jexp))
!                        (iexp,jexp) are the FITS image coordinates, which
!                         start counting from 1, just like (ipx, ipy) earlier.
!                        The integer values are the pixel centres, so that
!                         [iexp-0.5, iexp+0.5] x [jexp-0.5, jexp+0.5]
!                         are the bounds of the pixel.
                         if(iexp .gt. 0 .and. iexp .le. shape(0) .and. &
                            jexp .gt. 0 .and. jexp .le. shape(1)) then
                            fracx=min(xexpu,real(iexp)+0.5)-max(xexpl,real(iexp)-0.5)
                            fracy=min(yexpu,real(jexp)+0.5)-max(yexpl,real(jexp)-0.5)
                            frac=fracx*fracy ! /(xexpu-xexpl)**2 
                            ! we can't use Y here
                            ! since we want to account for different pixel
                            ! size in Y direction
                            do i=0,nband-1
!                              earr(i)%exp(iexp,jexp)=earr(i)%exp(iexp,jexp) + &
!                             Because CAL arrays count from 0,
!                              pixel coordinates (iexp,jexp) correspond
!                              to array index (iexp-1,jexp-1).
                              earr(i)%exp(iexp-1,jexp-1)=earr(i)%exp(iexp-1,jexp-1) + &
                                                   frac*fexpext(iccd,1)*area(i)
                            end do
                         end if
                      end do
                   end do
!                   if(notsaid) then
!                  No point in checking if it is gonna be .false. in any case.
                      notsaid=.false.
!                   end if
                   exit ! do ipy=ipyp,ipyp
                else ! (.not. useDetector)

                 if (fastmode) then
 
                  ! RDS
!                  ! rx0 = Xmm(ipy-1)/detypx2mm+pixoff
!                  ! ry0 = Ymm(ipy-1)/detypx2mm+pixoff
!                  rx0 = (Xmm(ipy-1)-opticsx_cc(0))/detypx2mm
!                  ry0 = (Ymm(ipy-1)-opticsy_cc(0))/detypx2mm
! Should not confuse between abs(detxpx2mm) and detypx2mm.
                  ! rx0 = Xmm(ipy-1)/abs(detxpx2mm)+pixoff
                  ! ry0 = Ymm(ipy-1)/abs(detypx2mm)+pixoff
                  rx0 = (Xmm(ipy-1)-opticsx_cc(0))/abs(detxpx2mm)
                  ry0 = (Ymm(ipy-1)-opticsy_cc(0))/abs(detypx2mm)

                  do lx=0,nsplit-1
                    rx=rx0+lx*pixfac
                    do ly=0,nsplit-1
                      ry=ry0+ly*pixfac
          
                      bix = (bmtrx10 * rx) + (bmtrx11 * ry) 
                      biy = (bmtrx20 * rx) + (bmtrx21 * ry)

                      do k=1,natt
                       if (fexpext(iccd,k)>0.0) then
 
                        xsky = detxcrpx + xnew(k) &
                               - (crnew(k)*bix - srnew(k)*biy)
                        ysky = detycrpx + ynew(k) &
                               - (srnew(k)*bix + crnew(k)*biy)
!                       apply empiric offsets:
!                        xsky=xsky+0.5        
!                        ysky=ysky-0.5


                        ix=xsky
                        iy=ysky
                        if (ix>0 .and. ix<naxis1-1 .and. &
                            iy>0 .and. iy<naxis2-1) then 
                          x0=ix
                          y0=iy
                          x1=ix+1 
                          y1=iy+1
                          xsp=xsky+fhx
                          xsn=xsky-fhx
                          ysp=ysky+fhy
                          ysn=ysky-fhy
                          if (xsn>x0 .and. xsp<x1 .and. &
                              ysn>y0 .and. ysp<y1) then 
                            do i=0,nband-1
                               wgh=area(i)*fexpext(iccd,k) 
                               if (wgh > 0.0) then
                                 exp1(i,k,ix,iy)=exp1(i,k,ix,iy)+fh2*wgh
                               end if
                            end do
                            exp0(k,ix,iy)=exp0(k,ix,iy)+fh2 
                          else
                            xm=min(xsp,x1)-max(xsn,x0)
                            ym=min(ysp,y1)-max(ysn,y0)
                            xp=max(xsp-x1,0.0)
                            yp=max(ysp-y1,0.0)
                            xn=max(x0-xsn,0.0)                    
                            yn=max(y0-ysn,0.0)
                            wmm=xm*ym
                            wpp=xp*yp
                            wmp=xm*yp
                            wnp=xn*yp
                            wnm=xn*ym
                            wnn=xn*yn         
                            wmn=xm*yn
                            wpn=xp*yn                          
                            wpm=xp*ym

                            do i=0,nband-1
                              wgh=area(i)*fexpext(iccd,k) 
                              if (wgh > 0.0) then
                                exp1(i,k,ix,iy)= &
                                exp1(i,k,ix,iy  )+wmm*wgh
                                exp1(i,k,ix-1,iy)= &
                                exp1(i,k,ix-1,iy)+wnm*wgh
                                exp1(i,k,ix+1,iy)= &
                                exp1(i,k,ix+1,iy)+wpm*wgh
                                if (yp > 0.0) then
                                  exp1(i,k,ix+1,iy+1)= &
                                  exp1(i,k,ix+1,iy+1)+wpp*wgh
                                  exp1(i,k,ix  ,iy+1)= &
                                  exp1(i,k,ix  ,iy+1)+wmp*wgh
                                  exp1(i,k,ix-1,iy+1)= &
                                  exp1(i,k,ix-1,iy+1)+wnp*wgh
                                end if
                                if (yn > 0.0) then
                                  exp1(i,k,ix-1,iy-1)= &
                                  exp1(i,k,ix-1,iy-1)+wnn*wgh
                                  exp1(i,k,ix  ,iy-1)= &
                                  exp1(i,k,ix  ,iy-1)+wmn*wgh
                                  exp1(i,k,ix+1,iy-1)= &
                                  exp1(i,k,ix+1,iy-1)+wpn*wgh
                                end if 
                              end if
                            end do
                            exp0(k,ix  ,iy  )=exp0(k,ix  ,iy  )+wmm
                            exp0(k,ix-1,iy  )=exp0(k,ix-1,iy  )+wnm
                            exp0(k,ix+1,iy  )=exp0(k,ix+1,iy  )+wpm
                            exp0(k,ix+1,iy+1)=exp0(k,ix+1,iy+1)+wpp
                            exp0(k,ix  ,iy+1)=exp0(k,ix  ,iy+1)+wmp
                            exp0(k,ix-1,iy+1)=exp0(k,ix-1,iy+1)+wnp
                            exp0(k,ix-1,iy-1)=exp0(k,ix-1,iy-1)+wnn
                            exp0(k,ix  ,iy-1)=exp0(k,ix  ,iy-1)+wmn
                            exp0(k,ix+1,iy-1)=exp0(k,ix+1,iy-1)+wpn
                          end if
                        end if
                       end if
                      end do ! k
                    end do
                  end do
     
                 else ! (.not. fastmode)

                    Xmm1(0)=Xmm(ipy-1)+xhsz_mm
                    Ymm1(0)=Ymm(ipy-1)-(yhsz_mm+yextra(ipy))
                    Zmm1(0)=Zmm(ipy-1)
                    do i=2,4
                       Xmm1(i-1)=Xmm1(0)-(1+(-1)**((i+1)/2))*xhsz_mm
!				= Xmm1(0) +  xhsz_mm              * [+0,-2,-2]
                       Ymm1(i-1)=Ymm1(0)+(1+(-1)**((i+2)/2))*(yhsz_mm+yextra(ipy))
!				= Ymm1(0) + (yhsz_mm+yextra(ipy)) * [+2,+2, 0]
                    end do
                    do k=1,natt
                     if (fexpext(iccd,k)>0.0) then
                       do i=1,4
!!                          rx = Xmm1(i-1)/ detypx2mm*80.
!!                          ry = Ymm1(i-1)/ detypx2mm*80.
!                          rx = (Xmm1(i-1)-opticsx_cc(0))/ detypx2mm*80.
!                          ry = (Ymm1(i-1)-opticsy_cc(0))/ detypx2mm*80.
! Should not confuse between abs(detxpx2mm) and detypx2mm.
!                          rx = Xmm1(i-1)/ abs(detxpx2mm)
!                          ry = Ymm1(i-1)/ abs(detypx2mm)
!                          rx = (Xmm1(i-1)-opticsx_cc(0))/ abs(detxpx2mm)
!                          ry = (Ymm1(i-1)-opticsy_cc(0))/ abs(detypx2mm)
! The formulae above are actually correct, but inconsistent with attcalc:
! 1. Calculations in attcalc are done in _integer_ DETX/Y, and the boresight
!    reference point is _rounded_ to integer DETX/Y before being referenced.
! 2. There is also a real bug in attcalc where int(... + 0.5d0) erroneously
!    adds +1 to all negative coordinates while attempting to round.
! 3. attcalc uses PLATE_SCALE / MM_PER_PIXEL for both MOS and PN,
!    while we have used PLATE_SCALE / MM_PER_PIXEL for PN but the
!    focal length formula for MOS above, to match e[mp]_events.
! The following hack bends (rx,ry) to match attcalc, to test for consistency.
                          rx = (Xmm1(i-1)- &
!	int(opticsx_cc(0)*CAL_getMiscellaneousDataValue('PLATE_SCALE_X')&
!                          /CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')&
!                          *20.0d0 + 0.5d0)&
	int(opticsx_cc(0)*rfac2x + 0.5d0)&
        * detxpx2mm / detxpx2mm_save / 180.0d0 / 72000.0d0 * pi * flngth &
				)/ abs(detxpx2mm)
                          ry = (Ymm1(i-1)- &
!	int(opticsy_cc(0)*CAL_getMiscellaneousDataValue('PLATE_SCALE_Y')&
!                          /CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y')&
!                          *20.0d0 + 0.5d0)&
        int(opticsy_cc(0)*rfac2y + 0.5d0)&
        * detypx2mm / detypx2mm_save / 180.0d0 / 72000.0d0 * pi * flngth &
				)/ abs(detypx2mm)
                          bix = (bmtrx10 * rx) + (bmtrx11 * ry) ! = Ysc 
                          biy = (bmtrx20 * rx) + (bmtrx21 * ry) ! = Zsc
!                          xsky4(i) = detxcrpx  + xnew(k)&
!                               - (crnew(k)*bix - srnew(k)*biy)/80.-0.5
!                          ysky4(i) = detycrpx  + ynew(k) &
!                               - (srnew(k)*bix + crnew(k)*biy)/80.-0.5
! Why would we want to shift the pixel corners by (-0.5,-0.5) in skycoord?
!  Because pix2map, called below,
! 1. treats integer coordinates as the *bottom-left corner* of a pixel, and
! 2. counts from (0,0) as opposed to (1,1) here.
! The net change needed to adapt to pix2map is a shift of (-0.5,-0.5).
! An offset of the same magnitude (but different sign for xsky) above is
!  described as empirical.  Why is this empirical?
                          xsky4(i) = detxcrpx  + xnew(k)&
                               - (crnew(k)*bix - srnew(k)*biy) - 0.5
                          ysky4(i) = detycrpx  + ynew(k) &
                               - (srnew(k)*bix + crnew(k)*biy) - 0.5
!                          xsky4(i) = detxcrpx  + xnew(k)&
!                               - (crnew(k)*bix - srnew(k)*biy)
!                          ysky4(i) = detycrpx  + ynew(k) &
!                               - (srnew(k)*bix + crnew(k)*biy)
! Note that for skycoords, det[xy]crpx is where the boresight is,
!  although the correctness of this behaviour depends on the implementation
!  of `evselect xcolumn=X ycolumn=Y keepimageset=Y`, and this dependence
!  does not seem to be documented anywhere.
                       end do
                       xsky4(0)=xsky4(4)
                       xsky4(5)=xsky4(1)
                       ysky4(0)=ysky4(4)
                       ysky4(5)=ysky4(1)
                       fk = sqrt(0.5*((xsky4(3)-xsky4(1))**2 + &
                                     (ysky4(3)-ysky4(1))**2))
! Why is there a factor of 0.5 above?  fk seems to be the sky solid angle.
                       pos=.false.
                       do i=0,nband-1
                         ! wght(i) = area(i)*transmission(i)*effgrad(i)* &
                         ! fexpext(iccd,k)
                         wgh = area(i)*fexpext(iccd,k)*fk**2
                         if (wgh > 0.0) then
                           wght(i)=wgh
                           pos=.true.
                         end if
                       end do 
                       if (pos) then

! Precalculation of resampling parameters to solve the speed issue. Alexis.
                         if(qalpha(k)) then
                            qalpha(k) = .false.
                            dx=xsky4(3)-xsky4(2)
                            dy=ysky4(3)-ysky4(2)
                            alpha = atan2(dy,dx)*180.d0/pi
                            if(alpha .lt. 0.d0) alpha=360.d0+alpha
                            alpha = alpha - 90.d0* (int(alpha)/90)
                               q45(k)=.false.
                            if(alpha .gt. 45.d0) then
                               q45(k)=.true.
                               alpha=90.d0 - alpha
                            endif
                            alpha=alpha*pi/180.d0
                            hs(k) = dsqrt(0.5d0)*dsin((pi/4.d0-alpha))
                            alcos(k) = dcos(alpha)
                            alsin(k) = dsin(alpha)
                         endif

                         call pix2map (xsky4,ysky4,wght,fk, &
                              naxis1, naxis2, nband, &
                              alsin(k),alcos(k),hs(k),q45(k))
                         ! call pix2map (xsky4,ysky4,wght,fk, &
                         !     naxis1, naxis2, nband)
                       end if ! (pos)
                     end if  ! (fexpext(iccd,k)>0.0)
                    end do ! k=1,natt

                 end if ! (.not. fastmode)

                end if ! (.not. useDetector)
             end do ! ipy
          end do ! ipyp
       end do ! ipx    
       deallocate(badpixtype)
      end if ! (ngti(iccd) > 0)
   
    end do ! iccd=nccd0, nccd
   
    if (.not. usedetector) then
     if (fastmode) then
      do i=0,nband-1
        earr(i)%exp=0.0
      end do
      exptest=-1.0
     end if
      do ix=0,naxis1-1
        do iy=0,naxis2-1
         if (fastmode)  then
          do k=1,natt 
            if (exp0(k,ix,iy)>0.0) then
              do i=0,nband-1
                earr(i)%exp(ix,iy)=earr(i)%exp(ix,iy) + &
                                   exp1(i,k,ix,iy)/exp0(k,ix,iy)
              end do
            end if
          end do
         end if
          if (earr(0)%exp(ix,iy)>exptest) then
            exptest=earr(0)%exp(ix,iy)
            xcen=ix
            ycen=iy
          end if
        end do
      end do
    end if

    do i=0,nband-1 
      call cp_head (set1, setO(i), "EPIC EXPOSURE MAP", 1)
      call setAttribute(setO(i),"XCEN",xcen,"X pos. of opt. axis")
      call setAttribute(setO(i),"YCEN",ycen,"Y pos. of opt. axis")
      call setAttribute(setO(i),"ATT_SRC","AHF     ","Source of attitude data (AHF|RAF|THF)")
      if (attinstr .eq. "EPN") then
        call setAttribute(setO(i),"OOTCORR",.TRUE.,"correction for point source OOT fraction applied?")
        call setAttribute(setO(i),"OOTFRAC",ootfrac,"OOT correction factor")
      endif
      call release(arr(i))
      call release(setO(i)) 
    end do

    if (nbb > 0) then
      call release (crawx)
      call release (crawy)
      call release (cye)
      call release (tabcoras)
    end if
    if (closeos) then 
! clean previous extension
!       call release (cox)
!       call release (cofs)
       call release (taboffs)
    end if

    call release (arr0)
    call release (set1)

    call release (set2)
    call release (set3)
!
! deallacote memory
!
    deallocate(earr)
    deallocate(fexpext)

  end subroutine eexpmap

!  subroutine atthst (ranomrad, sdecnom, cdecnom, skypixsize, time, ra, de, pa&
  subroutine atthst (ranomrad, sdecnom, cdecnom, skypixsizex, skypixsizey, time, ra, de, pa&
       , ncoras, gti, ngti, ngti_ext, nccds, set, xnew, ynew, crnew, srnew,&
       & npnt, eps, winx0, winy0, windx, windy, nextmax,ootfrac)

    type(start_stopT), dimension(:) :: gti 
    type(EulerAnglesType) :: euler
    real(kind=double), parameter  :: radian = 57.29577951d0
    real(kind=double), dimension(:) :: xnew, ynew, crnew, srnew
!    real(kind=double), dimension(:,:), allocatable :: fexpext
    real(kind=double) :: eps, fexp1, fexp0, fexp00 
    !integer(kind=int32), parameter :: nattmax=10000
    integer(kind=int32), dimension(12) :: winx0, winy0, windx, windy
    integer(kind=int32) :: ngti_ext, npnt, npnt1, ipnt, ncoras, j, iext, &
                           i, i0, nccds, nxx, iccd, nextmax
!    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsize
    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsizex, skypixsizey
    integer(kind=int32), dimension(:) :: ngti 
    integer(kind=int32), dimension(:), allocatable :: indx 
    real(kind=double), dimension(:), pointer :: time, ra, de, pa
    real(kind=double) :: rahelp, dechelp, relxb, relyb, rnb, rnew, sinr, cosr,&
         bor1, bor2, bor3, rn, relx, rely, relx2, rely2, diffPA, attPAxmm,&
         & attPArad, attdecrad, attrarad
    real(kind=double) :: tbeg, tend
    character(len=20) :: name
    character(len=6) :: xxname
    type(DataSetT) :: set
    type(TableT) :: tabcoras
    type(ColumnT) :: ctime,cfrac,ctimedel
    real(kind=single), dimension(:), pointer :: xxfrac
    real(kind=double), dimension(:), pointer :: xxtime
    real(kind=single), dimension(:), pointer :: xxtimedel
    real(kind=single) :: pntimedel,ootfrac,frametim
!!!!! Needed for gfortran to pass the correct array descriptor to C++
#ifdef NAGf90Fortran
    real(kind=double), dimension(:,:), pointer :: Bmatrix
#else
    real(kind=double), dimension(:,:), pointer :: Bmatrix
   real(kind=double), dimension(3,3) :: tmpBmatrix
#endif
    allocate (indx(1:ncoras))

! 1) analyze attitude changes

    npnt=1
    do i0=0,ncoras-1
      if (dabs(ra(i0)).le.360d0 .and. &
          dabs(de(i0)).le. 90d0 .and. &
          dabs(pa(i0)).le.360d0) then
        indx(npnt)=i0
        exit
      end if
    end do
    i0=i0+1 
    if (i0.gt.1) call warning("NullValues", &
                 "NULL values in the attitude table were ignored") 

    do i=i0,ncoras-1

      if (dabs(ra(i)).le.360d0 .and. &
          dabs(de(i)).le. 90d0 .and. &
          dabs(pa(i)).le.360d0) then

       if (   (dabs(ra(i) - ra(indx(npnt)))) .gt. eps .or. &
            & (dabs(de(i) - de(indx(npnt)))) .gt. eps .or. &
!            & (dabs(pa(i) - pa(indx(npnt)))) .gt. 200.0*eps) then
            & (dabs(pa(i) - pa(indx(npnt)))) .gt. eps) then
! The following are the criteria used in attcalc:
!       if (   (dabs(ra(i) - ra(indx(npnt)))) .gt. (1.0d-7 * radian) .or. &
!            & (dabs(de(i) - de(indx(npnt)))) .gt. (1.0d-7 * radian) .or. &
!            & (dabs(pa(i) - pa(indx(npnt)))) .gt. (1.0d-7 * radian)) then
! Gotta be consistent to prevent events from going out of bound.
! Thus, to produce sky exposure maps matching events exactly,
! users should call eexpmap withdetcoords=no attrebin=0.020626481 (or smaller).
! 0.020626481 arcsec == 1e-7 radian
! We point this out in the eexpmap document.
          npnt=npnt+1
          !if (npnt > nattmax) call error("BufferFull", &
          !              "More than 10000 attitude bins")
          indx(npnt)=i
       end if
      else
        call warning("NullValues", &
                     "NULL values in the attitude table were ignored")        
      end if
    end do

    allocate (fexpext(nextmax,npnt))    
    xxname = "EXPOSU"
   
! 2) merge with GTI to create the vector of factors

    do iccd=1,nccds

! set  default window ranges in case no exposure extension is present
       winx0(iccd)=0
       winy0(iccd)=0
       windx(iccd)=999
       windy(iccd)=999 

       write(name,"(i2.2)") iccd  
       name=xxname//name(1:2)
       if (hasBlock(set,name)) then          
          tabcoras = table(set,name)

! get window ranges
          if (hasAttribute(tabcoras,"WINDOWX0")) then
            winx0(iccd)=real32Attribute(tabcoras,"WINDOWX0")
          else
            winx0(iccd)=0
          end if
          if (hasAttribute(tabcoras,"WINDOWY0")) then
            winy0(iccd)=real32Attribute(tabcoras,"WINDOWY0")
          else
            winy0(iccd)=0
          end if
          if (hasAttribute(tabcoras,"WINDOWDX")) then
            windx(iccd)=real32Attribute(tabcoras,"WINDOWDX")
          else
            windx(iccd)=999
          end if
          if (hasAttribute(tabcoras,"WINDOWDY")) then
            windy(iccd)=real32Attribute(tabcoras,"WINDOWDY")
          else
            windy(iccd)=999
          end if

          nxx = numberOfRows(tabcoras)
          ctime = column(tabcoras,"TIME", Read)    
          xxtime => real64Data(ctime)
          if (hasColumn(tabcoras,"TIMEDEL")) then
            ctimedel=column(tabcoras,"TIMEDEL", Read)
            xxtimedel => real32Data(ctimedel)
          else if (hasAttribute(tabcoras,"TIMEDEL")) then
              pntimedel=real32Attribute(tabcoras,"TIMEDEL")
              allocate(xxtimedel(0:size(xxtime)-1))
!              write(*,*) 'size(xxtime)-1)',size(xxtime)-1
              xxtimedel = pntimedel 
!              write(*,*) 'timdel',pntimedel, xxtimedel(0), xxtimedel(1), xxtimedel(1000)
              if (hasAttribute(tabcoras,"FRAMETIM")) then
                 frametim=real32Attribute(tabcoras,"FRAMETIM")
                 if (frametim .ne. 0) then
                   ootfrac=pntimedel/frametim*1000.
                 endif
              endif
          endif
          cfrac = column(tabcoras,"FRACEXP", Read)
          xxfrac => real32Data(cfrac)
       write(errstr,*) iccd, ' extension XX: N points is',nxx
       call message(AppMsg,VERBOSE)

       iext = min(ngti_ext, iccd) ! case of similar GTI for all chips

       do i=1,npnt
          fexpext(iccd,i)=0.
          do j=0,ngti(iext)-1
             if(gti(iext)%stop(j) .gt. time(indx(i))) then
                if (i == npnt) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(gti(iext)%stop(j),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxtimedel, xxfrac, nxx)
!                   write(*,*) 'fexpext(iccd,i)',iccd,i,fexpext(iccd,i)
                else if (i == 1) then
                   tbeg = gti(iext)%start(j)
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxtimedel, xxfrac, nxx)
!                   write(*,*) 'fexpext(iccd,i)',iccd,i,fexpext(iccd,i)
                else if (gti(iext)%start(j) .lt. time(indx(i+1))) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxtimedel, xxfrac, nxx) 
!                   write(*,*) 'fexpext(iccd,i)',iccd,i,fexpext(iccd,i)                  
                end if
             end if
          end do
! exposure extension is accounted for by calling xxcti
!          write (errstr,*) i,iccd,fexpext(iccd,i)
!          call message(AppMsg,VERBOSE)
       end do

       else
          nxx=0
          call warning("NoExpoExt","Exposure extension not found")
       iext = min(ngti_ext, iccd) ! case of similar GTI for all chips
       do i=1,npnt                                      
          fexpext(iccd,i)=0.
          do j=0,ngti(iext)-1
             if(gti(iext)%stop(j) .gt. time(indx(i))) then
                if (i == npnt) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(gti(iext)%stop(j),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                else if (i == 1) then
                   tbeg = gti(iext)%start(j)
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                else if (gti(iext)%start(j) .lt. time(indx(i+1))) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                end if
             end if
          end do
! exposure extension is accounted for by calling xxcti
!          write (errstr,*) i,iccd,fexpext(iccd,i)
!          call message(AppMsg,VERBOSE)
       end do
       end if

       if (hasBlock(set,name)) then
          call release (ctime)
          call release (cfrac)
          call release (tabcoras)
       end if
    end do

! 2b) identify time bins with similar attitude    

!    For each time bin ipnt, assign to fexp00 the maximum exposure time
!     (for that time bin only) amongst the CCDs, fexpext(iccd,ipnt).
!    Then, assign to fexp0 the sum of fexp00 from all time bins,
!     and multiply to result by 1.0d-4.
!    Note that this is the sum of exposure time of possibly different CCDs
!     in different time bins.
     npnt1=npnt
if (.false.) then
     fexp0=0d0
     do ipnt=1,npnt
       fexp00=0d0
       do iccd=1,nccds
         fexp00=max(fexp00,fexpext(iccd,ipnt))
       end do
       fexp0=fexp0+fexp00
     end do 
     fexp0=fexp0*1d-4

!    Check each time bin to see if the per-bin exposure time of all CCDs
!     in that bin are shorter than fexp0 from above (ie, < 0.1% of a quantity
!     at least as big as the total exposure time of the entire exposure).
!    If all are shorter than 0.1%, then ignore the attitude information
!     as well as the exposure times for this time bin.
     do ipnt=npnt,1,-1
       fexp1=0d0
       do iccd=1,nccds
         fexp1=max(fexp1,fexpext(iccd,ipnt))
       end do
       if (fexp1<fexp0) then
         npnt1=npnt1-1
         do j=ipnt,npnt1
           do iccd=1,nccds
             fexpext(iccd,j)=fexpext(iccd,j+1)
           end do          
           indx(j)=indx(j+1)
         end do
       end if
     end do
end if
! We cannot actually ignore short time bins as above (within the ad-hoc
!  if (.false.) then ... end if, or else events incident at those times
!  can go to image pixels of zero exposure.

! Group time bins with attitude differences below the tolerance together
!  by adding all the exposure times together, assigning it to the first
!  such bin, and removing the other bins.
! This multi-level do-loop can be rewritten to speed things up.
     do ipnt=npnt1,1,-1
       do i=ipnt-1,1,-1
         if ( (dabs(ra(indx(i)) - ra(indx(ipnt)))) < eps .and. &
              (dabs(de(indx(i)) - de(indx(ipnt)))) < eps .and. &
!              (dabs(pa(indx(i)) - pa(indx(ipnt)))) < 200.0*eps) then
              (dabs(pa(indx(i)) - pa(indx(ipnt)))) < eps) then
! The following are the criteria used in attcalc:
!         if ( (dabs(ra(indx(i)) - ra(indx(ipnt)))) < (1.0d-7 * radian) .and. &
!              (dabs(de(indx(i)) - de(indx(ipnt)))) < (1.0d-7 * radian) .and. &
!              (dabs(pa(indx(i)) - pa(indx(ipnt)))) < (1.0d-7 * radian)) then
! Gotta be consistent to prevent events from going out of bound.
! Thus, to produce sky exposure maps matching events exactly,
! users should call eexpmap withdetcoords=no attrebin=0.020626481 (or smaller).
! 0.020626481 arcsec == 1e-7 radian
! We point this out in the eexpmap document.
           do iccd=1,nccds
             fexpext(iccd,i)=fexpext(iccd,i)+fexpext(iccd,ipnt)
           end do          
           npnt1=npnt1-1
           do j=ipnt,npnt1
             do iccd=1,nccds
               fexpext(iccd,j)=fexpext(iccd,j+1)
             end do          
             indx(j)=indx(j+1)
           end do
           exit   
         end if
       end do
     end do
     npnt=npnt1

    write (errstr,*) 'number of att-histogram bins ', npnt
    call message(AppMsg,VERBOSE)

! 3) get transformation to sky coordinates

    do i=1,npnt
       
       attrarad = ra(indx(i))/radian
       attdecrad = de(indx(i))/radian
       attPArad = pa(indx(i))/radian

       attPAxmm = 4.712388980 - attPArad  
       call spher2tang (attrarad, attdecrad, ranomrad, sdecnom, cdecnom, &
            relx, rely)
    !--------------------------------------------------------------!
    !	do same 1deg away local north (help point) (SUBR) Ysc-axis !
    !--------------------------------------------------------------!
       call helppoint (attrarad, attdecrad, attPAxmm, & 
            rahelp, dechelp)
       call spher2tang (rahelp, dechelp, ranomrad , sdecnom, cdecnom, & 
            relx2, rely2)
!print*, 'attrarad=',attrarad, attdecrad, attPAxmm, rahelp, dechelp,relx,rely,relx2,rely2

    !-----------------------!     
    !	PA correction       ! 
    !-----------------------!
       diffPA = datan2(relx-relx2, rely2-rely)
       rn = sngl(4.71238898 - diffPA) !270 - diffpa; rn now in apa sense
       if (rn > 6.28319) rn = rn - 6.283185307
       if (rn < 0.0) rn = rn + 6.283185307

    !------------------------------------------!
    ! 	get boresight (SIAM) Bmatrix and angles         
    !------------------------------------------!
!       Bmatrix => CAL_toDirectionCosineMatrix(CAL_getBoresight( &
!                                                 time(indx(i))))
! attcalc is still using the old formula:
        call CAL_getBoresightMatrix(time(indx((i))), Bmatrix)
!               CAL_getBoresightMatrix(time(indx(i)), Bmatrix)
#ifdef NAGf90Fortran
    call OAL_toEulerAngles(Bmatrix, euler)
#else
!!!!! Needed for gfortran to pass the correct array descriptor to C++
!!!!! If this local variable is not created, the array descriptor we
!!!!! pass to C++ is not correct
    tmpBmatrix = Bmatrix
!!!!! Then, we have to transpose the matrix to preserve the order of the
!!!!! elements in C++. I know, it is ugly.
!!!!! This issue is also present in rgslib package.
    call OAL_toEulerAngles(transpose(Bmatrix), euler)
#endif
        !-----------------------------------------------------------!
        !  convert to bor3, bor2, bor1  - for the moment, hardwire  !
        !-----------------------------------------------------------!
       bor3 = euler%phi
       bor2 = euler%theta
       bor1 = euler%psi
       call CAL_releaseMemory(Bmatrix)
       bor3 = dsin(bor3)
       bor2 = dsin(bor2)
       bor1 = dsin(bor1)
! test - no boresight
!        bor3 = 0.0 !dsin(pi)
!        bor2 = 0.0 !dsin(0.0)
!        bor1 = 0.0 !dsin(0.0)
        !---------------------------------------------!
        !   apply boresight correction (SUBR) to att  !
        !---------------------------------------------!
       sinr = dsin(dble(rn))
       cosr = dcos(dble(rn))
       call borxy (bor3, bor2, bor1, relx, rely, rn, relxb, relyb, rnb, &
            sinr, cosr)
       rnew = rnb * radian			!rnew in degrees in apa sense
       srnew(i) = sin(rnb)			!i.e. on sky anticlock (celN)
       crnew(i) = cos(rnb)			!xnew, ynew in x/y frame
!       xnew(i) = relxb * radian/skypixsize	!all b/s corrected
!       ynew(i) = relyb * radian/skypixsize 	!
       xnew(i) = relxb * radian/skypixsizex	!all b/s corrected
       ynew(i) = relyb * radian/skypixsizey	!
   end do

  end subroutine atthst

!---------------------------------------------------------------------
!
  function xxcti(tbeg, tend, xxtime, xxtimedel, xxfrac, nxx)
!
!---------------------------------------------------------------------
! eXposure extension Corrected Time Interval
    real(kind=double) :: tbeg, tend, sum, xxcti
    integer(kind=int32) :: nxx, nin, i
    real(kind=single), dimension(:), pointer :: xxfrac
    real(kind=single), dimension(:), pointer :: xxtimedel
    real(kind=double), dimension(:), pointer :: xxtime
    

    xxcti = tend - tbeg
          
    if ((.not. xxcti>0.) .or. (nxx == 0)) return


    nin = 0
    sum = 0.
    do i = 0, nxx-1
       if(xxtime(i) .ge. tbeg .and. xxtime(i) .le. tend) then
          nin=nin+1
          sum=sum+xxfrac(i)*xxtimedel(i)
       end if
    end do

    if (nin == 0) return
       
!    xxcti = xxcti * sum / nin
    xxcti=sum     

  end function xxcti

!---------------------------------------------------------------------






  real function get_BadRadius(ibbc)
    !Event pattern selection of ibbc (higher value includes lower values)
    !  5 4 3 4 5
    !  4 2 1 2 4
    !  3 1 0 1 3
    !  4 2 1 2 4
    !  5 4 3 4 5
    !
    ! For PN valid grades up to 2. For MOS up to 5.

    real(kind=single), dimension (0:5) :: radius
    integer(kind=int32), intent(inout) :: ibbc

    if(ibbc .gt. 5) then
       write(errstr,*) 'Not valid grade for cleaning! Must be < 5'
       call message(AppMsg,Verbose)
       ibbc = 5
    end if

    radius(0)=0.
    radius(1)=1.0
    radius(2)=1.42
    radius(3)=2.0
    radius(4)=2.24
    radius(5)=2.83


    get_BadRadius = radius(ibbc)+0.01

  end function get_BadRadius


!---------------------------------------------------------------------
!  logical function wmode (iccd, nsubmode, ipx, ipy)
!
!  commented out 05/01/2004
!
!  integer, intent(in) :: iccd, nsubmode
!  integer(kind=int16), intent(in) :: ipx, ipy
!
!  wmode = .true.
!
!  if (nsubmode >= 15 .and. ipy <= 15) wmode=.false. 
!
!  if (nsubmode == 17) then
!    if (ipy < 101) wmode = .false.
!  else if (nsubmode == 18) then
!    if (iccd /= 4 .or. iccd == 4 .and. ipy < 137) wmode = .false.
!  else if (nsubmode == 2 .or. nsubmode == 4) then
!    if (iccd == 1 .and. &
!       (ipx < 251 .or. ipx > 350 .or. & 
!        ipy < 251 .or. ipy > 350)) wmode = .false.
!  else if (nsubmode == 3 .or. nsubmode == 5) then
!    if (iccd == 1 .and. &
!       (ipx < 151 .or. ipx > 450 .or. & 
!        ipy < 151 .or. ipy > 450)) wmode = .false.
!  end if
!
!  end function wmode 
!---------------------------------------------------------------------


!---------------------------------------------------------------------
!
!	*** SUBROUTINES from attcalc
!
!---------------------------------------------------------------------
    subroutine spher2tang (attra, attdec, ranom, sdecnom, cdecnom, relx, &
                           rely)
    implicit none
    real(kind=double)                     :: attra, attdec, sdecnom, &
                                        cdecnom, relx, rely, &
                                        ranom, diffra, dsinra, &
                                        dcosra, dsindec, dcosdec, &
                                        d1s, dh, d1c, dranew, d2s, d2c, &
                                        ddecnew

    diffra = attra - ranom
    dsinra = dsin(diffra)
    dcosra = dcos(diffra)
    dsindec = dsin(attdec)
    dcosdec = dcos(attdec)

!	residual alpha dranew
    d1s = dsinra * dcosdec
    dh = - dcosdec * dcosra
    d1c = sdecnom*dsindec - dh*cdecnom
    dranew = datan2(d1s,d1c)

! 	residual delta ddecnew
    d2s = dh*sdecnom + dsindec*cdecnom
    d2c = dsqrt(1.0d0 - d2s*d2s)
    ddecnew = datan2(d2s,d2c)

! 	convert to x,y in NS-system
    relx = - dranew
    rely =  - ddecnew
    end subroutine spher2tang
  !---------------------------------------------------------------------
    subroutine helppoint (attra, attdec, attPA, &
                           rahelp, dechelp)

    implicit none
    real(kind=single), parameter        :: sin1deg = 1.74524064E-2, &
         cos1deg = 9.99847695E-1
    real(kind=double)           :: attra, attdec, attPA, &
                              rasin, decsin, PAsin, racos, deccos, &
                              PAcos, xhelp, yhelp, &
                              zhelp, rhelp, rahelp, &
                              dechelp

!	calc sines and cosines
      rasin = dsin(attra)
      racos = dcos(attra)
      decsin = dsin(attdec)
      deccos = dcos(attdec)
      PAsin = dsin(attPA)
      PAcos = dcos(attPA)

!	cartesian co-ords of help point
       xhelp = (sin1deg*(PAcos*racos*decsin - PAsin*rasin)) + &
                        (cos1deg*racos*deccos)
       yhelp = (sin1deg*(PAsin*racos + PAcos*rasin*decsin)) + &
                        (cos1deg*rasin*deccos)
       
       zhelp = - sin1deg*PAcos*deccos + cos1deg*decsin

!	length squared of help vector
      rhelp = xhelp*xhelp + yhelp*yhelp +zhelp*zhelp

      if (rhelp < 0.d0) then
!	error
        call warning("undefinedHelpVector","Help vector has undefined length - taking aspect solution as help point to contine")
        rahelp = attra
        dechelp = attdec
      else

!	spherical co-ords of help point
        rhelp = dsqrt(rhelp)
        rahelp = datan2 (yhelp, xhelp)
          if (rahelp < 0.d0) rahelp = rahelp + (2*3.1415926536)
        dechelp = dasin(zhelp/rhelp)
      endif

    end subroutine helppoint
  !---------------------------------------------------------------------
    subroutine borxy (bor3, bor2, bor1, relx, rely, rn, relxb, &
                      relyb, rnb, sinr, cosr)

    implicit none
    real(kind=double)       :: bor3, bor2, bor1, relx, rely, rn, relxb, &
                            relyb, rnb, sinr, cosr, xbor, ybor

    xbor = 0.0
    ybor = 0.0
    xbor = (cosr * bor3) + (sinr * bor2) !in x/y frame
    ybor = (sinr * bor3) - (cosr * bor2) !in x/y frame

!       shift attitude vector by (xbor,ybor)

    relxb = relx + xbor
    relyb = - rely + ybor                !minus to put rely in x/y frame

!       correct "corrected Pos.Angle" also by fibor
! RDS CHANGE THIS SIGN OR NOT ??
!    rnb = rn + bor1   - ORIGINAL RETURN TO THIS
    rnb = rn - bor1  ! correct version - use when psi is correct in CCF


    end subroutine borxy

!-----------------------------------------------------------------
 subroutine merge_gti (lgti, ngti, fstart, fstop, irange, itfilt, itfilt1)

 real(kind=double), dimension(0:gtimax), intent(IN) :: fstart, fstop 
 integer, intent(IN) :: irange, itfilt
 integer, intent(INOUT) :: ngti
 
 integer :: j,k, n, m, ntest, mtest, itfilt1
 real(kind=double) :: aux
 real(kind=double), dimension(2*(ngti+irange)) :: tbuff
! real(kind=double), dimension(0:ngti+irange-1), target :: tstart, tstop 
 logical, dimension(2*(ngti+irange)) :: type
 logical :: bux
 
 type(start_stopT) :: lgti

!   itfilt1=itfilt+ngti-1

   k=1
   do j=0,ngti-1
     tbuff(k)=lgti%start(j)
     type(k)=.true. 
     k=k+1
     tbuff(k)=lgti%stop(j)
     type(k)=.false.
     k=k+1
   end do 
   do j=0,irange-1
     tbuff(k)=fstart(j)
     type(k)=.true.
     k=k+1
     tbuff(k)=fstop(j)
     type(k)=.false.
     k=k+1
   end do

   do n=1,k-1
     do m=n+1,k-1
       if (tbuff(n)>tbuff(m)) then
         aux=tbuff(n) 
         tbuff(n)=tbuff(m)
         tbuff(m)=aux
         bux=type(n)
         type(n)=type(m)
         type(m)=bux
       end if
     end do
   end do 


   if (ngti>0) then
     deallocate (lgti%start)
     deallocate (lgti%stop)
   end if 
   allocate (lgti%start(0:ngti+irange-1)) 
   allocate (lgti%stop(0:ngti+irange-1)) 

   ntest=0
   j=0 
   do n=1,k-1
     mtest=ntest
     if (type(n)) then
       ntest=ntest+1
     else
       ntest=ntest-1
     end if
     if (ntest == itfilt1+1 .and. mtest == itfilt1) then
       lgti%start(j) = tbuff(n)
!       tstart(j) = tbuff(n)
     end if
     if (ntest == itfilt1 .and. mtest == itfilt1+1) then 
       lgti%stop(j) = tbuff(n)
!       tstop(j) = tbuff(n)
       j=j+1
     end if
   end do

   ngti = j

!   gti(i)%start => tstart
!   gti(i)%stop => tstop
  
 end subroutine merge_gti
  
!------------------------------------------------------------------------------------------

 subroutine get_margins(instr,iccd,flagmin,flagmax,flagmask,nmask,flagmargins,flagmargins2,bordermargin,offsetmargin)

!  use flag selections from DSS keywords to define 
!  margins for CCD borders, bad pixels, and offset columns    

   character(len=20), intent(in) :: instr
   integer(kind=int32), intent(in), dimension(0:100) :: flagmin, flagmax 
   integer(kind=int32), intent(in), dimension(0:100,0:100) :: flagmask
   integer(kind=int32), intent(in) :: iccd, nmask
   integer(kind=int32), intent(out), dimension(1:100) :: flagmargins, flagmargins2
   integer(kind=int32), intent(out) :: bordermargin, offsetmargin
   integer(kind=int32) :: imask


   ! check for "close to border" flag:
   bordermargin=0
   if (instr == "EPN") then
     if (flagmax(iccd) < 4) bordermargin = 1
     do imask=0,nmask-1
        if (btest(flagmask(iccd,imask),2)) bordermargin = 1
     end do
   endif

   if (instr == "EMOS1" .or. instr == "EMOS2") then
     if (flagmax(iccd) < 2) bordermargin = 1
     do imask=0,nmask-1
        if (btest(flagmask(iccd,imask),1)) bordermargin = 1
     end do
   endif

   ! check for "next to offset column" flag:
   offsetmargin=0
   if (instr == "EPN") then
     if (flagmax(iccd) < 16) offsetmargin = 1
     do imask=0,nmask-1
        if (btest(flagmask(iccd,imask),4)) offsetmargin = 1
     end do
   endif

   ! check for "close to onboard bad pix" flag:
     flagmargins(1)=0
     flagmargins2(1)=0
     if (flagmax(iccd) < 32) then
       flagmargins(1) = 1
       flagmargins2(1) = 1
     endif
     do imask=0,nmask-1
       if (btest(flagmask(iccd,imask),5)) then
         flagmargins(1)=1
         flagmargins2(1)=1         
       endif
     end do


   ! check for "close to dead pix flag (GL 09/08/2006):
     flagmargins(2)=0
      flagmargins2(2)=0
      if (flagmax(iccd) < 256) then
        flagmargins(2) = 1
        flagmargins2(2) = 1
      endif
      do imask=0,nmask-1
        if (btest(flagmask(iccd,imask),8)) then
          flagmargins(2)=1
          flagmargins2(2)=1         
        endif
      end do
 

   ! check for "close to onboard bright pix" flag:
     flagmargins(3)=0
     flagmargins2(3)=0
     if (flagmax(iccd) < 64) then
       flagmargins(3) = 1
       flagmargins2(3) = 1
     endif
     do imask=0,nmask-1
       if (btest(flagmask(iccd,imask),6)) then
         flagmargins(3)=1
         flagmargins2(3)=1         
       endif
     end do
 

 end subroutine get_margins

!------------------------------------------------------------------------------------------
end module eexpmap_mod  

